from __future__ import annotations

import re
from collections import OrderedDict
from itertools import permutations as perm

from .. import types as lgdo

_lgdo_datatype_map: dict[str, lgdo.LGDO] = OrderedDict(
    [
        (lgdo.Scalar, r"^real$|^bool$|^complex$|^bool$|^string$"),
        (lgdo.VectorOfVectors, r"^array<1>\{array<1>\{.+\}\}$"),
        (lgdo.VectorOfEncodedVectors, r"^array<1>\{encoded_array<1>\{.+\}\}$"),
        (
            lgdo.ArrayOfEncodedEqualSizedArrays,
            r"^array_of_encoded_equalsized_arrays<1,1>\{.+\}$",
        ),
        (
            lgdo.Histogram,
            rf"^struct\{{(?:{'|'.join([','.join(p) for p in perm(['binning', 'weights', 'isdensity'])])})\}}$",
        ),
        (lgdo.Struct, r"^struct\{.*\}$"),
        (lgdo.Table, r"^table\{.*\}$"),
        (lgdo.FixedSizeArray, r"^fixedsize_array<\d+>\{.+\}$"),
        (lgdo.ArrayOfEqualSizedArrays, r"^array_of_equalsized_arrays<1,1>\{.+\}$"),
        (lgdo.Array, r"^array<\d+>\{.+\}$"),
    ]
)
"""Mapping between LGDO types and regular expression defining the corresponding datatype string"""


def datatype(expr: str) -> type:
    """Return the LGDO type corresponding to a datatype string."""
    expr = expr.strip()
    for type_, regex in _lgdo_datatype_map.items():
        if re.search(regex, expr):
            return type_

    msg = f"unknown datatype '{expr}'"
    raise RuntimeError(msg)


def get_nested_datatype_string(expr: str) -> str:
    """Matches the content of the outermost curly brackets."""
    return re.search(r"\{(.*)\}$", expr).group(1)


def get_struct_fields(expr: str) -> list[str]:
    """Returns a list of :class:`~.lgdo.types.struct.Struct` fields, given its datatype string."""
    fields = get_nested_datatype_string(expr).split(",")
    return [] if fields == [""] else fields



from __future__ import annotations

import awkward as ak
import numpy as np
import pandas as pd
import pint
import pytest

import lgdo


def test_datatype_name():
    aoesa = lgdo.ArrayOfEqualSizedArrays()
    assert aoesa.datatype_name() == "array_of_equalsized_arrays"


def test_form_datatype():
    aoesa = lgdo.ArrayOfEqualSizedArrays(dims=(2, 3))
    assert aoesa.form_datatype() == "array_of_equalsized_arrays<2,3>{real}"


def test_init():
    attrs = {"attr1": 1}
    aoesa = lgdo.ArrayOfEqualSizedArrays(
        dims=(2, 3), dtype=np.float32, fill_val=42, attrs=attrs
    )
    assert aoesa.dims == (2, 3)
    assert (aoesa.nda == np.full((2, 3), 42, np.float32)).all()
    assert aoesa.attrs == attrs | {"datatype": "array_of_equalsized_arrays<2,3>{real}"}


def test_to_vov():
    aoesa = lgdo.ArrayOfEqualSizedArrays(
        nda=np.array([[53, 91, 66, 58, 8], [78, 57, 66, 88, 73], [85, 99, 86, 68, 53]])
    )
    vov = aoesa.to_vov()
    assert isinstance(vov, lgdo.VectorOfVectors)
    assert np.array_equal(vov[0], [53, 91, 66, 58, 8])
    assert np.array_equal(vov[1], [78, 57, 66, 88, 73])
    assert np.array_equal(vov[2], [85, 99, 86, 68, 53])

    vov = aoesa.to_vov([2, 5, 6])
    assert np.array_equal(vov[0], [53, 91])
    assert np.array_equal(vov[1], [78, 57, 66])
    assert np.array_equal(vov[2], [85])


def test_view():
    aoesa = lgdo.ArrayOfEqualSizedArrays(
        nda=np.array([[53, 91, 66, 58, 8], [78, 57, 66, 88, 73], [85, 99, 86, 68, 53]]),
        attrs={"units": "m"},
    )

    v = aoesa.view_as("np", with_units=True)
    assert isinstance(v, pint.Quantity)
    assert v.u == "meter"
    assert np.array_equal(v.m, aoesa.nda)

    v = aoesa.view_as("np", with_units=False)
    assert isinstance(v, np.ndarray)

    v = aoesa.view_as("pd", with_units=False)
    assert isinstance(v, pd.Series)

    v = aoesa.view_as("ak", with_units=False)
    assert isinstance(v, ak.Array)

    with pytest.raises(ValueError):
        aoesa.view_as("pd", with_units=True)

    v = aoesa.view_as("ak", with_units=True)
    assert isinstance(v, ak.Array)
    assert ak.parameters(v) == {"units": "m"}



from __future__ import annotations

import logging
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

import hist
import numpy as np
import pandas as pd
from numpy.typing import NDArray

from .array import Array
from .lgdo import LGDO
from .scalar import Scalar
from .struct import Struct

log = logging.getLogger(__name__)


class Histogram(Struct):
    class Axis(Struct):
        def __init__(
            self,
            edges: NDArray | Array | None,
            first: float | None,
            last: float | None,
            step: float | None,
            closedleft: bool = True,
            binedge_attrs: dict[str, Any] | None = None,
        ) -> None:
            """
            A special struct to group axis parameters for use in a :class:`Histogram`.

            Depending on the parameters, an axis either can have

            * a binning described by a range object, if ``first``, ``last`` and ``step``
              are passed, or
            * a variable binning described by the ``edges`` array.

            Parameters
            ----------
            edges
                an array of edges that describe the binning of this axis.
            first
                left edge of the leftmost bin
            last
                right edge of the rightmost bin
            step
                step size (width of each bin)
            closedleft
                if True, the bin intervals are left-closed :math:`[a,b)`;
                if False, intervals are right-closed :math:`(a,b]`.
            binedge_attrs
                attributes that will be added to the ``binedges`` LGDO that
                is part of the axis struct.
            """
            if edges is not None and (
                first is not None or last is not None or step is not None
            ):
                msg = "can only construct Axis either from edges or from range"
                raise ValueError(msg)
            if edges is None and (first is None or last is None or step is None):
                msg = "did not pass all range parameters"
                raise ValueError(msg)

            if edges is None:
                edges = Struct(
                    {
                        "first": Scalar(first),
                        "last": Scalar(last),
                        "step": Scalar(step),
                    },
                    binedge_attrs,
                )
            else:
                if not isinstance(edges, Array):
                    edges = Array(edges, attrs=binedge_attrs)
                elif binedge_attrs is not None:
                    msg = "passed both binedge as Array LGDO instance and binedge_attrs"
                    raise ValueError(msg)

                if len(edges.nda.shape) != 1:
                    msg = "must pass an array<1>{real} as edges vector"
                    raise ValueError(msg)

            super().__init__({"binedges": edges, "closedleft": Scalar(closedleft)})

        @classmethod
        def from_edges(
            cls,
            edges: NDArray | Iterable[float],
            binedge_attrs: dict[str, Any] | None = None,
        ) -> Histogram.Axis:
            """Create a new axis with variable binning described by ``edges``."""
            edges = np.array(edges)
            return cls(edges, None, None, None, True, binedge_attrs)

        @classmethod
        def from_range_edges(
            cls,
            edges: NDArray | Iterable[float],
            binedge_attrs: dict[str, Any] | None = None,
        ) -> Histogram.Axis:
            """Create a new axis from the binning described by ``edges``, but try to convert it to
            a evenly-spaced range object first.

            .. warning ::

                This function might return a wrong binning, especially in the case of very small
                magnitudes of the spacing. See the documentation of :func:`numpy.isclose` for
                details. Use this function only with caution, if you know the binning's order of
                magniutude.
            """
            edges = np.array(edges)
            edge_diff = np.diff(edges)
            if np.any(~np.isclose(edge_diff, edge_diff[0])):
                return cls(edges, None, None, None, True, binedge_attrs)
            return cls(None, edges[0], edges[-1], edge_diff[0], True, binedge_attrs)

        @property
        def is_range(self) -> bool:
            return isinstance(self["binedges"], Struct)

        @property
        def first(self) -> float:
            if not self.is_range:
                msg = "Axis is not a range"
                raise TypeError(msg)
            return self["binedges"]["first"].value

        @property
        def last(self) -> float:
            if not self.is_range:
                msg = "Axis is not a range"
                raise TypeError(msg)
            return self["binedges"]["last"].value

        @property
        def step(self) -> float:
            if not self.is_range:
                msg = "Axis is not a range"
                raise TypeError(msg)
            return self["binedges"]["step"].value

        @property
        def closedleft(self) -> bool:
            return self["closedleft"].value

        @property
        def nbins(self) -> int:
            """Return the number of bins, both for variable and range binning."""
            if self.is_range:
                bins = (self.last - self.first) / self.step
                bins_int = int(np.rint(bins))
                assert np.isclose(bins, bins_int)
                return bins_int
            return len(self["binedges"].nda) - 1

        @property
        def edges(self) -> NDArray:
            """Return all binedges, both for variable and range binning."""
            if self.is_range:
                return np.linspace(self.first, self.last, self.nbins + 1)
            return self["binedges"].nda

        def __str__(self) -> str:
            thr_orig = np.get_printoptions()["threshold"]
            np.set_printoptions(threshold=8)

            if self.is_range:
                string = f"first={self.first}, last={self.last}, step={self.step}"
            else:
                string = f"edges={self.edges}"
            string += f", closedleft={self.closedleft}"

            attrs = self.get_binedgeattrs()
            if attrs:
                string += f" with attrs={attrs}"

            np.set_printoptions(threshold=thr_orig)
            return string

        def get_binedgeattrs(self, datatype: bool = False) -> dict:
            """Return a copy of the LGDO attributes dictionary of the binedges

            Parameters
            ----------
            datatype
                if ``False``, remove ``datatype`` attribute from the output
                dictionary.
            """
            return self["binedges"].getattrs(datatype)

    def __init__(
        self,
        weights: hist.Hist | NDArray | Array,
        binning: None
        | Iterable[Histogram.Axis]
        | Iterable[NDArray]
        | Iterable[tuple[float, float, float]] = None,
        isdensity: bool = False,
        attrs: dict[str, Any] | None = None,
        binedge_attrs: dict[str, Any] | None = None,
        flow: bool = True,
    ) -> None:
        """A special struct to contain histogrammed data.

        Parameters
        ----------
        weights
            An :class:`numpy.ndarray` to be used for this object's internal
            array, or a :class:`hist.Hist` object, whose data view is used
            for this object's internal array.
            Note: the array/histogram view is used directly, not copied
        binning
            * has to by None if a :class:`hist.Hist` has been passed as ``weights``
            * can be a list of pre-initialized :class:`Histogram.Axis`
            * can be a list of tuples, each representing a range, ``(first, last, step)``
            * can be a list of numpy arrays, as returned by :func:`numpy.histogramdd`.
        isdensity
            If True, all bin contents represent a density (amount per volume), and not
            an absolute amount.
        binedge_attrs
            attributes that will be added to the all ``binedges`` of all axes.
            This does not work if :class:`Histogram.Axis` instances are directly passed
            as binning.
        attrs
            a set of user attributes to be carried along with this LGDO.
        flow
            If ``False``, discard counts in over-/underflow bins of the passed
            :class:`hist.Hist` instance. If ``True``, this data will also be discarded,
            but a warning is emitted.

            .. note ::

                :class:`Histogram` does not support storing counts in overflow or
                underflow bins. This parameter just controls, whether a warning will
                be emitted.
        """
        if isinstance(weights, hist.Hist):
            if binning is not None:
                msg = "not allowed to pass custom binning if constructing from hist.Hist instance"
                raise ValueError(msg)
            if isdensity:
                msg = "not allowed to pass isdensity=True if constructing from hist.Hist instance"
                raise ValueError(msg)

            if weights.sum(flow=True) != weights.sum(flow=False) and flow:
                log.warning(
                    "flow bins of hist.Hist cannot be represented, their counts are discarded"
                )
            weights_view = weights.view(flow=False)
            if type(weights_view) is not np.ndarray:
                msg = "only simple numpy-backed storages can be used in a hist.Hist"
                raise ValueError(msg)
            w = Array(weights_view)

            b = []
            for ax in weights.axes:
                if not isinstance(ax, (hist.axis.Regular, hist.axis.Variable)):
                    msg = "only regular or variable axes of hist.Hist can be converted"
                    raise ValueError(msg)
                if isinstance(ax, hist.axis.Regular):
                    step = (ax.edges[-1] - ax.edges[0]) / ax.size
                    bax = Histogram.Axis(
                        None, ax.edges[0], ax.edges[-1], step, True, binedge_attrs
                    )
                    b.append(bax)
                else:
                    b.append(Histogram.Axis.from_edges(ax.edges, binedge_attrs))
        else:
            if binning is None:
                msg = "need to pass binning to construct Histogram"
                raise ValueError(msg)

            # set up binning
            if all(isinstance(ax, Histogram.Axis) for ax in binning):
                if binedge_attrs is not None:
                    msg = "passed both binedges as Axis instances and binedge_attrs"
                    raise ValueError(msg)
                b = binning
            elif all(isinstance(ax, np.ndarray) for ax in binning):
                b = [Histogram.Axis.from_edges(ax, binedge_attrs) for ax in binning]
            elif all(isinstance(ax, tuple) for ax in binning):
                b = [Histogram.Axis(None, *ax, True, binedge_attrs) for ax in binning]
            else:
                msg = "invalid binning object passed"
                raise ValueError(msg)

            # set up bin weights
            if isinstance(weights, Array):
                w = weights
            elif weights is None:
                w = Array(shape=[ax.nbins for ax in b], fill_val=0, dtype=np.float32)
            else:
                w = Array(weights)

            if len(binning) != len(w.nda.shape):
                msg = "binning and weight dimensions do not match"
                raise ValueError(msg)
            for i, ax in enumerate(b):
                if ax.nbins != w.nda.shape[i]:
                    msg = f"bin count does not match weight count along axis {i}"
                    raise ValueError(msg)

        b = Struct({f"axis_{i}": a for i, a in enumerate(b)})

        super().add_field("binning", b)
        super().add_field("weights", w)
        super().add_field("isdensity", Scalar(isdensity))
        LGDO.__init__(self, attrs)

    @property
    def isdensity(self) -> bool:
        return self["isdensity"].value

    @property
    def weights(self) -> Array:
        return self["weights"]

    @property
    def binning(self) -> tuple[Histogram.Axis, ...]:
        bins = sorted(self["binning"].items())
        assert all(isinstance(v, Histogram.Axis) for k, v in bins)
        return tuple(v for _, v in bins)

    def fill(self, data, w: NDArray = None, keys: Sequence[str] = None) -> None:
        """Fill histogram by incrementing bins with data points weighted by w

        Parameters
        ----------
        data
            a ndarray with inner dimension equal to number of axes, or a list
            of equal-length 1d-arrays containing data for each axis, or a
            Mapping to 1d-arrays containing data for each axis (requires keys),
            or a Pandas dataframe (optionally takes a list of keys)
        w
            weight to use for incrementing data points. If None, use 1 for all
        keys
            list of keys to use if data is a pandas ''DataFrame'' or ''Mapping''
        """
        if keys is not None:
            if isinstance(keys, str):
                keys = [keys]
            elif not isinstance(keys, list):
                keys = list(keys)

        if (
            isinstance(data, np.ndarray)
            and len(data.shape) == 1
            and len(self.binning) == 1
        ):
            N = len(data)
            data = [data]
        elif (
            isinstance(data, np.ndarray)
            and len(data.shape) == 2
            and data.shape[1] == len(self.binning)
        ):
            N = data.shape[0]
            data = data.T
        elif isinstance(data, pd.DataFrame) and (
            (keys is not None and len(keys) == len(self.binning))
            or data.ndim == len(self.binning)
        ):
            if keys is not None:
                data = data[keys]
            N = len(data)
            data = data.values.T
        elif isinstance(data, Sequence) and len(data) == len(self.binning):
            data = [d if isinstance(d, np.ndarray) else np.array(d) for d in data]
            N = len(data[0])
            if not all(len(d) == N for d in data):
                msg = "length of all data arrays must be equal"
                raise ValueError(msg)
        elif isinstance(data, Mapping):
            if not isinstance(keys, Sequence) or len(keys) != len(self.binning):
                msg = "filling hist with Mapping data requires a list of keys with same length as histogram rank"
                raise ValueError(msg)
            data = [
                data[k] if isinstance(data[k], np.ndarray) else np.array(data[k])
                for k in keys
            ]
            N = len(data[0])
            if not all(len(d) == N for d in data):
                msg = "length of all data arrays must be equal"
                raise ValueError(msg)
        else:
            msg = "data must be 2D numpy array or list of 1D arrays with length equal to number of axes"
            raise ValueError(msg)

        idx = np.zeros(N, np.float64)  # bin indices for flattened array
        oor_mask = np.ones(N, np.bool_)  # mask to remove out of range values
        stride = [s // self.weights.dtype.itemsize for s in self.weights.nda.strides]
        for col, ax, s in zip(data, self.binning, stride, strict=False):
            if ax.is_range:
                idx += s * np.floor((col - ax.first) / ax.step - int(not ax.closedleft))
                if ax.closedleft:
                    oor_mask &= (ax.first <= col) & (col < ax.last)
                else:
                    oor_mask &= (ax.first < col) & (col <= ax.last)
            else:
                idx += s * (
                    np.searchsorted(
                        ax.edges, col, side=("right" if ax.closedleft else "left")
                    )
                    - 1
                )
                if ax.closedleft:
                    oor_mask &= (ax.edges[0] <= col) & (col < ax.edges[-1])
                else:
                    oor_mask &= (ax.edges[0] < col) & (col <= ax.edges[-1])

        # increment bin contents
        idx = idx[oor_mask].astype(np.int64)
        w = w[oor_mask] if w is not None else 1
        np.add.at(self.weights.nda.reshape(-1), idx, w)

    def __setitem__(self, name: str, obj: LGDO) -> None:
        # do not allow for new attributes on this
        known_keys = ("binning", "weights", "isdensity")
        if name in known_keys and name not in self:
            # but allow initialization while unpickling (after __init__() this is unreachable)
            Struct.__setitem__(self, name, obj)
        else:
            msg = "histogram fields cannot be mutated "
            raise AttributeError(msg)

    def __getattr__(self, name: str) -> None:
        # do not allow for new attributes on this
        msg = "histogram fields cannot be mutated"
        raise AttributeError(msg)

    def add_field(self, name: str | int, obj: LGDO) -> None:  # noqa: ARG002
        """
        .. error ::

            Not applicable: A histogram cannot be used as a struct
        """
        msg = "histogram fields cannot be mutated"
        raise TypeError(msg)

    def remove_field(self, name: str | int, delete: bool = False) -> None:  # noqa: ARG002
        """
        .. error ::

            Not applicable: A histogram cannot be used as a struct
        """
        msg = "histogram fields cannot be mutated"
        raise TypeError(msg)

    def __str__(self) -> str:
        string = "{\n"
        for k, v in enumerate(self.binning):
            string += f" 'axis_{k}': {v},\n"
        string += "}"

        attrs = self.getattrs()
        if attrs:
            string += f" with attrs={attrs}"

        return string

    def view_as(
        self,
        library: str,
    ) -> tuple[NDArray] | hist.Hist:
        r"""View the histogram data as a third-party format data structure.

        This is typically a zero-copy or nearly zero-copy operation.

        Supported third-party formats are:

        - ``np``: returns a tuple of binning and an :class:`np.ndarray`, similar
          to the return value of :func:`numpy.histogramdd`.
        - ``hist``: returns an :class:`hist.Hist` that holds **a copy** of this
          histogram's data.

        Warning
        -------
        Viewing as ``hist`` will perform a copy of the stored histogram data.

        Parameters
        ----------
        library
            format of the returned data view.

        See Also
        --------
        .LGDO.view_as
        """
        if library == "hist":
            if self.isdensity:
                msg = "hist.Hist cannot represent density histograms"
                raise ValueError(msg)

            hist_axes = []
            for a in self.binning:
                if not a.closedleft:
                    msg = "hist.Hist cannot represent right-closed intervals"
                    raise ValueError(msg)
                if a.is_range:
                    hist_ax = hist.axis.Regular(
                        bins=a.nbins,
                        start=a.first,
                        stop=a.last,
                        underflow=False,
                        overflow=False,
                    )
                else:
                    hist_ax = hist.axis.Variable(
                        a.edges,
                        underflow=False,
                        overflow=False,
                    )
                hist_axes.append(hist_ax)

            return hist.Hist(*hist_axes, data=self.weights.view_as("np"))

        if library == "np":
            edges = tuple([a.edges for a in self.binning])
            return self.weights.view_as("np"), edges

        msg = f"{library!r} is not a supported third-party format."
        raise TypeError(msg)



from __future__ import annotations

import numpy as np

from lgdo import ArrayOfEncodedEqualSizedArrays, ArrayOfEqualSizedArrays
from lgdo.compression import varlen

rng = np.random.default_rng()


def _to_bin(array):
    return [np.binary_repr(x, width=8) for x in array]


def test_zigzag():
    assert varlen.zigzag_encode(0) == 0
    assert varlen.zigzag_encode(-1) == 1
    assert varlen.zigzag_encode(1) == 2
    assert varlen.zigzag_encode(-2) == 3
    assert varlen.zigzag_encode(2) == 4

    assert varlen.zigzag_decode(np.uint8(0)) == 0
    assert varlen.zigzag_decode(np.uint8(1)) == -1
    assert varlen.zigzag_decode(np.uint8(2)) == 1
    assert varlen.zigzag_decode(np.uint8(3)) == -2
    assert varlen.zigzag_decode(np.uint8(4)) == 2


def test_varint_encoding():
    # >>> import varint
    # >>> ["{0:08b}".format(b) for b in varint.encode(x)]
    expected = {
        1: ["00000001"],
        5: ["00000101"],
        127: ["01111111"],
        128: ["10000000", "00000001"],
        150: ["10010110", "00000001"],
        7896: ["11011000", "00111101"],
        8192: ["10000000", "01000000"],
        268435455: ["11111111", "11111111", "11111111", "01111111"],
    }

    for x, varint in expected.items():
        encx = np.empty(10, dtype="ubyte")
        pos = varlen.uleb128_encode(x, encx)
        assert _to_bin(encx[:pos]) == varint
        assert varlen.uleb128_decode(encx[:pos]) == (x, len(varint))


def test_varint_encode_decode_equality():
    for x in [1, 3856, 234, 11, 93645, 2, 45]:
        encx = np.empty(100, dtype="ubyte")
        pos = varlen.uleb128_encode(x, encx)
        assert varlen.uleb128_decode(encx[:pos])[0] == x


def test_uleb128zzdiff_encode_decode_equality():
    sig_in = np.array([1, 3856, 234, 11, 93645, 2, 45], dtype="uint32")
    sig_out = np.empty(100, dtype="ubyte")
    nbytes = np.empty(1, dtype="uint32")

    varlen.uleb128_zigzag_diff_array_encode(sig_in, sig_out, nbytes)

    encx = np.empty(10, dtype="ubyte")
    offset = 0
    last = 0
    for s in sig_in:
        pos = varlen.uleb128_encode(varlen.zigzag_encode(int(s) - last), encx)
        assert np.array_equal(sig_out[offset : offset + pos], encx[:pos])
        offset += pos
        last = int(s)

    sig_in_dec = np.empty(100, dtype="uint32")
    siglen = np.empty(1, dtype="uint32")
    varlen.uleb128_zigzag_diff_array_decode(sig_out, nbytes, sig_in_dec, siglen)
    assert np.array_equal(sig_in, sig_in_dec[0 : siglen[0]])


def test_uleb128zzdiff_encode_decode_2dmatrix_equality():
    sig_in = rng.integers(0, 10000, (10, 1000), dtype="uint32")
    sig_out = np.empty((10, 1000 * 4), dtype="ubyte")
    nbytes = np.empty(10, dtype="uint32")

    varlen.uleb128_zigzag_diff_array_encode(sig_in, sig_out, nbytes)

    sig_in_dec = np.empty((10, 1000), dtype="uint32")
    siglen = np.empty(10, dtype="uint32")
    varlen.uleb128_zigzag_diff_array_decode(sig_out, nbytes, sig_in_dec, siglen)

    assert np.array_equal(sig_in, sig_in_dec)


def test_uleb128zzdiff_encode_decode_wrapper_2dmatrix_equality():
    sig_in = rng.integers(0, 10000, (10, 1000), dtype="uint32")
    sig_in_dec, siglen = varlen.decode(varlen.encode(sig_in))
    assert np.array_equal(siglen, np.full(10, 1000))

    assert np.array_equal(sig_in, sig_in_dec[:, : siglen[0]])

    sig_in_dec = np.empty((10, 1000), dtype="uint32")
    varlen.decode(varlen.encode(sig_in), sig_in_dec)

    assert np.array_equal(sig_in, sig_in_dec)


def test_uleb128zzdiff_encode_decode_lgdo_aoesa(wftable):
    voev = varlen.encode(wftable.values)
    assert isinstance(voev, ArrayOfEncodedEqualSizedArrays)
    assert voev.decoded_size.value == wftable.wf_len

    for i in range(len(voev)):
        wf_enc, nbytes = varlen.encode(wftable.values[i])
        assert len(voev[i]) == nbytes
        assert np.array_equal(voev[i], wf_enc[:nbytes])
        wf_dec, siglen = varlen.decode((wf_enc, nbytes))
        assert np.array_equal(wf_dec[:siglen], wftable.values[i])

    assert (
        voev.encoded_data.to_aoesa(fill_val=0, preserve_dtype=True).nda.dtype
        == np.ubyte
    )

    sig_in_dec = varlen.decode(voev)
    assert isinstance(sig_in_dec, ArrayOfEqualSizedArrays)
    assert sig_in_dec == wftable.values

    # decode with pre-allocated LGDO
    sig_in_dec = ArrayOfEqualSizedArrays(
        dims=(1, 1), shape=(len(wftable), wftable.wf_len), dtype="uint16"
    )
    varlen.decode(voev, sig_in_dec)
    assert sig_in_dec == wftable.values


def test_encoding_decoding_empty():
    aoesa = ArrayOfEqualSizedArrays(dims=(1, 1), nda=np.empty((0, 0)))
    enc_aoesa = varlen.encode(aoesa)
    assert enc_aoesa == ArrayOfEncodedEqualSizedArrays()
    assert varlen.decode(enc_aoesa) == aoesa



from __future__ import annotations

from pathlib import Path

import numpy as np

from lgdo import lh5, types
from lgdo.lh5 import utils


def test_expand_path_with_path_objects(tmp_path):
    f = tmp_path / "file.lh5"
    f.touch()
    # direct Path
    assert utils.expand_path(f) == str(f)
    # Path with base_path as Path
    assert utils.expand_path(f.name, base_path=tmp_path) == f.name
    assert utils.expand_path(Path(f.name), base_path=Path(tmp_path)) == f.name
    # wildcard with Path
    assert utils.expand_path(tmp_path / "*.lh5", list=True) == [str(f)]


def test_read_write_with_path_objects(tmp_path):
    arr = types.Array(np.arange(5))
    out = tmp_path / "out.lh5"
    lh5.write(arr, "arr", out, group="/data", wo_mode="overwrite_file")
    result = lh5.read("/data/arr", out)
    assert np.array_equal(result.nda, np.arange(5))


def test_store_with_path_objects(lh5_file, tmp_path):
    store = lh5.LH5Store(base_path=tmp_path)
    path_obj = Path(lh5_file)
    obj = store.read("/data/struct/scalar", path_obj)
    assert obj.value == 10
    out = tmp_path / "out_store.lh5"
    arr = types.Array(np.arange(3))
    store.write(arr, "arr", out, group="/data", wo_mode="overwrite_file")
    obj2 = store.read("/data/arr", out)
    assert np.array_equal(obj2.nda, np.arange(3))



"""legend-pydataobj's command line interface utilities."""

from __future__ import annotations

import argparse
import logging
import sys

from . import __version__, lh5
from . import logging as lgdogging  # eheheh
from .lh5.concat import lh5concat

log = logging.getLogger(__name__)


def lh5ls(args=None):
    """:func:`.lh5.show` command line interface."""
    parser = argparse.ArgumentParser(
        prog="lh5ls", description="Inspect LEGEND HDF5 (LH5) file contents"
    )

    # global options
    parser.add_argument(
        "--version",
        action="store_true",
        help="""Print legend-pydataobj version and exit""",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )

    parser.add_argument(
        "lh5_file",
        help="""Input LH5 file""",
    )
    parser.add_argument("lh5_group", nargs="?", help="""LH5 group.""", default="/")
    parser.add_argument(
        "--attributes", "-a", action="store_true", help="""Print HDF5 attributes too"""
    )
    parser.add_argument(
        "--depth",
        "-d",
        type=int,
        default=None,
        help="""Maximum tree depth of groups to print""",
    )
    parser.add_argument(
        "--detail",
        action="store_true",
        help="""Print details about datasets""",
    )

    args = parser.parse_args(args)

    if args.verbose:
        lgdogging.setup(logging.DEBUG, logging.getLogger("lgdo"))
    elif args.debug:
        lgdogging.setup(logging.DEBUG, logging.root)
    else:
        lgdogging.setup()

    if args.version:
        print(__version__)  # noqa: T201
        sys.exit()

    lh5.show(
        args.lh5_file,
        args.lh5_group,
        attrs=args.attributes,
        depth=args.depth,
        detail=args.detail,
    )


def lh5concat_cli(args=None):
    """Command line interface for concatenating array-like LGDOs in LH5 files."""
    parser = argparse.ArgumentParser(
        prog="lh5concat",
        description="""
Concatenate LGDO Arrays, VectorOfVectors and Tables in LH5 files.

Examples
--------

Concatenate all eligible objects in file{1,2}.lh5 into concat.lh5:

  $ lh5concat -o concat.lh5 file1.lh5 file2.lh5

Include only the /data/table1 Table:

  $ lh5concat -o concat.lh5 -i /data/table1/* file1.lh5 file2.lh5

Exclude the /data/table1/col1 Table column:

  $ lh5concat -o concat.lh5 -e /data/table1/col1 file1.lh5 file2.lh5
        """,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # global options
    parser.add_argument(
        "--version",
        action="store_true",
        help="""Print legend-pydataobj version and exit""",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )

    parser.add_argument(
        "lh5_file",
        nargs="+",
        help="""Input LH5 files""",
    )
    parser.add_argument(
        "--output",
        "-o",
        help="""Output file""",
        default="lh5concat-output.lh5",
    )
    parser.add_argument(
        "--overwrite",
        "-w",
        action="store_true",
        help="""Overwrite output file""",
    )
    parser.add_argument(
        "--include",
        "-i",
        help="""Regular expression (fnmatch style) for object names that should
        be concatenated. To include full tables, you need to explicitly include
        all its columns with e.g. '/path/to/table/*'. The option can be passed
        multiple times to provide a list of patterns.
        """,
        action="append",
        default=None,
    )
    parser.add_argument(
        "--exclude",
        "-e",
        help="""List of object names that should be excluded. Takes priority
        over --include. See --include help for more details.
        """,
        action="append",
        default=None,
    )

    args = parser.parse_args(args)

    if args.verbose:
        lgdogging.setup(logging.INFO, logging.getLogger("lgdo"))
    elif args.debug:
        lgdogging.setup(logging.DEBUG, logging.root)
    else:
        lgdogging.setup()

    if args.version:
        print(__version__)  # noqa: T201
        sys.exit()

    lh5concat(
        lh5_files=args.lh5_file,
        overwrite=args.overwrite,
        output=args.output,
        include_list=args.include,
        exclude_list=args.exclude,
    )



from __future__ import annotations

from warnings import warn

import numpy as np

from . import types as lgdo
from .lh5 import utils


def copy(obj: lgdo.LGDO, dtype: np.dtype = None) -> None:
    warn(
        "lgdo_utils.copy will soon be removed and will be replaced soon with copy member functions of each LGDO data type.",
        DeprecationWarning,
        stacklevel=2,
    )
    return utils.copy(obj, dtype)


def get_element_type(obj: object) -> str:
    warn(
        "'lgdo.lgdo_utils' has been renamed to 'lgdo.utils'. "
        "Please replace either 'import lgdo.lgdo_utils as utils' with 'import lgdo.utils as utils' "
        "or 'from lgdo.lgdo_utils import get_element_type' with 'from lgdo.utils import get_element_type'."
        "'lgdo.lgdo_utils' will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return utils.get_element_type(obj)


def parse_datatype(datatype: str) -> tuple[str, tuple[int, ...], str | list[str]]:
    warn(
        "'lgdo.lgdo_utils' has been renamed to 'lgdo.utils'. "
        "Please replace either 'import lgdo.lgdo_utils as utils' with 'import lgdo.utils as utils' "
        "or 'from lgdo.lgdo_utils import parse_datatype' with 'from lgdo.utils import parse_datatype'."
        "'lgdo.lgdo_utils' will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return utils.parse_datatype(datatype)


def expand_vars(expr: str, substitute: dict[str, str] | None = None) -> str:
    warn(
        "'lgdo.lgdo_utils' has been renamed to 'lgdo.utils'. "
        "Please replace either 'import lgdo.lgdo_utils as utils' with 'import lgdo.utils as utils' "
        "or 'from lgdo.lgdo_utils import expand_vars' with 'from lgdo.utils import expand_vars'."
        "'lgdo.lgdo_utils' will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return utils.expand_vars(expr, substitute)


def expand_path(
    path: str,
    substitute: dict[str, str] | None = None,
    list: bool = False,
    base_path: str | None = None,
) -> str | list:
    warn(
        "'lgdo.lgdo_utils' has been renamed to 'lgdo.utils'. "
        "Please replace either 'import lgdo.lgdo_utils as utils' with 'import lgdo.utils as utils' "
        "or 'from lgdo.lgdo_utils import expand_path' with 'from lgdo.utils import expand_path'."
        "'lgdo.lgdo_utils' will be removed in a future release. ",
        DeprecationWarning,
        stacklevel=2,
    )
    return utils.expand_path(path, substitute, list, base_path)



"""
Implements a LEGEND Data Object representing a struct and corresponding
utilities.
"""

from __future__ import annotations

import copy
import logging
import re
from collections.abc import Iterable, Iterator, Mapping, MutableMapping
from itertools import chain
from typing import Any

import numpy as np

from .lgdo import LGDO

log = logging.getLogger(__name__)
# parse for name1/name2 or name1.name2
parser = re.compile("([^\\./]*)(?:[\\./](.+))?")


class Struct(LGDO, MutableMapping):
    """A dictionary of LGDO's with an optional set of attributes.

    After instantiation, add fields using :meth:`add_field` to keep the
    datatype updated, or call :meth:`update_datatype` after adding.
    """

    def __new__(cls, *args, **kwargs) -> Struct:
        obj = super().__new__(cls, *args, **kwargs)
        obj.obj_dict = {}
        return obj

    def __init__(
        self,
        obj_dict: Mapping[str, LGDO] | None = None,
        attrs: Mapping[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        obj_dict
            instantiate this Struct using the supplied named LGDO's.  Note: no
            copy is performed, the objects are used directly.
        attrs
            a set of user attributes to be carried along with this LGDO.
        """
        if obj_dict is not None:
            self.update(obj_dict)

        # check the datatype attribute passed by the user and sort the fields
        # to ensure consistent behavior
        if attrs is not None and "datatype" in attrs:
            _attrs = copy.copy(dict(attrs))

            if not _is_struct_datatype(self.datatype_name(), _attrs["datatype"]):
                msg = (
                    f"datatype attribute ({self.attrs['datatype']}) is not "
                    f"compatible with class datatype!"
                )
                raise ValueError(msg)

            _attrs["datatype"] = _sort_datatype_fields(_attrs["datatype"])
            attrs = _attrs

        super().__init__(attrs)

    def datatype_name(self) -> str:
        return "struct"

    def form_datatype(self) -> str:
        return (
            self.datatype_name()
            + "{"
            + ",".join(sorted([str(k) for k in self.keys()]))
            + "}"
        )

    def update_datatype(self) -> None:
        self.attrs["datatype"] = self.form_datatype()

    def add_field(self, name: str | int, obj: LGDO | Mapping[str, LGDO]) -> None:
        """Add a field to the table or set an existing field.

        Parameters
        ----------
        name
            key to use for field. Key can be nested (e.g. ``name1.name2`` or
            ``name1/name2``); this will navigate through the tree, creating
            new fields as needed
        obj
            object to add. Can be any LGDO object, or a mapping from names
            to LGDO objects that will be converted to an LGDO :class:`.Struct`
        """
        name1, name2 = parser.match(name).groups()
        if name2:
            if not name1:
                self.add_field(name2, obj)
            else:
                if name1 not in self:
                    self.add_field(name1, Struct())
                self[name1].add_field(name2, obj)
        else:
            if not isinstance(obj, LGDO):
                if isinstance(obj, Mapping):
                    obj = Struct(obj)
                else:
                    msg = f"value of '{name}' ({obj!r}) is not an LGDO or a Mapping"
                    raise ValueError(msg)

            self.obj_dict[name1] = obj
            self.update_datatype()

    def __getitem__(self, name: str) -> LGDO:
        """Get value associated with field. Name can be nested (e.g. ``name1.name2``
        or ``name1/name2``); this will search in nested Structs
        """
        name1, name2 = parser.match(name).groups()
        obj = self.obj_dict[name1] if name1 else self
        return obj if not name2 else obj[name2]

    def __setitem__(self, name: str, obj: LGDO) -> None:
        return self.add_field(name, obj)

    def __delitem__(self, name: str) -> None:
        self.remove_field(name, delete=True)

    def __iter__(self) -> Iterator[str]:
        "Iterator over top level fields"
        return iter(self.obj_dict)

    def __len__(self) -> int:
        "Return number of fields"
        return len(self.obj_dict)

    def update(
        self, other: Struct | Mapping[str, LGDO] | Iterable[str, LGDO] = (), /, **kwargs
    ) -> None:
        """Add or set a field(s) to the table or set an existing field. For
        nested Structs, only update at the lowest level of nesting; unlike for
        nested dicts, nested fields not included in other will not be removed.

        Parameters
        ----------
        other
            Struct/Mapping from fields to new values
        """
        for k, v in chain(
            other.items() if isinstance(other, Mapping) else other, kwargs.items()
        ):
            if isinstance(v, Mapping) and k in self:
                self[k].update(v)
            else:
                self[k] = v

    # Note: in principle these are automatically defined by Mapping; however, they will get the length wrong for Tables
    def keys(self):
        return self.obj_dict.keys()

    def items(self):
        return self.obj_dict.items()

    def values(self):
        return self.obj_dict.values()

    def __getattr__(self, name: str) -> LGDO:
        if name in self.obj_dict:
            return self.obj_dict[name]

        raise AttributeError(name)

    def remove_field(self, name: str | int, delete: bool = False) -> None:
        """Remove a field from the table.

        Parameters
        ----------
        name
            name of the field to be removed.
        delete
            if ``True``, delete the field object by calling :any:`del`.
        """
        name1, name2 = parser.match(name).groups()
        if name2:
            if not name1:
                self.remove_field(name2, delete)
            else:
                self[name1].remove_field(name2, delete)

        else:
            if delete:
                del self.obj_dict[name1]
            else:
                self.pop(name1)
            self.update_datatype()

    def __str__(self) -> str:
        """Convert to string (e.g. for printing)."""

        thr_orig = np.get_printoptions()["threshold"]
        np.set_printoptions(threshold=8)

        string = "{\n"
        for k, v in self.items():
            if "\n" in str(v):
                rv = str(v).replace("\n", "\n    ")
                string += f" '{k}':\n    {rv},\n"
            else:
                string += f" '{k}': {v},\n"
        string += "}"

        attrs = self.getattrs()
        if attrs:
            string += f" with attrs={attrs}"

        np.set_printoptions(threshold=thr_orig)

        return string

    def __repr__(self) -> str:
        npopt = np.get_printoptions()
        np.set_printoptions(threshold=5, edgeitems=2, linewidth=100)
        out = (
            self.__class__.__name__
            + "(dict="
            + self.obj_dict.__repr__()
            + f", attrs={self.attrs!r})"
        )
        np.set_printoptions(**npopt)
        return " ".join(out.replace("\n", " ").split())

    def view_as(self) -> None:
        r"""View the Struct data as a third-party format data structure.

        Error
        -----
        Not implemented. Since Struct's fields can have different lengths,
        converting to a NumPy, Pandas or Awkward is generally not possible.
        Call :meth:`.LGDO.view_as` on the fields instead.

        See Also
        --------
        .LGDO.view_as
        """
        msg = (
            "Since Struct's fields can have different lengths, "
            "converting to a NumPy, Pandas or Awkward is generally "
            "not possible. Call view_as() on the fields instead."
        )
        raise NotImplementedError(msg)


def _is_struct_datatype(dt_name, expr):
    return re.search("^" + dt_name + r"\{(.*)\}$", expr) is not None


def _get_struct_fields(expr: str) -> list[str]:
    assert _is_struct_datatype(".*", expr)

    arr = re.search(r"\{(.*)\}$", expr).group(1).split(",")
    if arr == [""]:
        arr = []

    return sorted(arr)


def _struct_datatype_equal(dt_name, dt1, dt2):
    if any(not _is_struct_datatype(dt_name, dt) for dt in (dt1, dt2)):
        return False

    return _get_struct_fields(dt1) == _get_struct_fields(dt2)


def _sort_datatype_fields(expr):
    assert _is_struct_datatype(".*", expr)

    match = re.search(r"^(.*)\{.*\}$", expr)
    struct_type = match.group(1)
    fields = _get_struct_fields(expr)

    return struct_type + "{" + ",".join(sorted([str(k) for k in fields])) + "}"



from __future__ import annotations

import shutil
import uuid
from getpass import getuser
from pathlib import Path
from tempfile import gettempdir

import pytest
from legendtestdata import LegendTestData

_tmptestdir = Path(gettempdir()) / f"lgdo-tests-{getuser()}-{uuid.uuid4()!s}"


@pytest.fixture(scope="session")
def tmptestdir():
    Path(_tmptestdir).mkdir(parents=True, exist_ok=True)
    return _tmptestdir


def pytest_sessionfinish(session, exitstatus):  # noqa: ARG001
    if exitstatus == 0:
        shutil.rmtree(_tmptestdir, ignore_errors=True)


@pytest.fixture(scope="session")
def lgnd_test_data():
    ldata = LegendTestData()
    ldata.checkout("df10dde")
    return ldata



from __future__ import annotations

import logging
import sys

import h5py
import numba
import numpy as np

from ....types import (
    Array,
    VectorOfVectors,
)
from ... import datatype as dtypeutils
from ...exceptions import LH5DecodeError
from .array import (
    _h5_read_array,
)
from .utils import read_attrs

log = logging.getLogger(__name__)


def _h5_read_vector_of_vectors(
    h5g,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    obj_buf=None,
    obj_buf_start=0,
):
    if obj_buf is not None and not isinstance(obj_buf, VectorOfVectors):
        msg = "object buffer is not a VectorOfVectors"
        raise LH5DecodeError(msg, fname, oname)

    # read out cumulative_length
    cumulen_buf = None if obj_buf is None else obj_buf.cumulative_length
    h5d_cl = h5py.h5d.open(h5g, b"cumulative_length")
    cumulative_length, n_rows_read = _h5_read_array(
        h5d_cl,
        fname,
        f"{oname}/cumulative_length",
        start_row=start_row,
        n_rows=n_rows,
        idx=idx,
        use_h5idx=use_h5idx,
        obj_buf=cumulen_buf,
        obj_buf_start=obj_buf_start,
    )
    # get a view of just what was read out for cleaner code below
    this_cumulen_nda = cumulative_length.nda[
        obj_buf_start : obj_buf_start + n_rows_read
    ]

    if idx is not None and n_rows_read > 0:
        # get the starting indices for each array in flattened data:
        # the starting index for array[i] is cumulative_length[i-1]
        idx2 = np.asarray(idx).copy() - 1

        # re-read cumulative_length with these indices
        # note this will allocate memory for fd_starts!
        fd_start = None
        if idx2[0] == -1:
            idx2 = idx2[1:]
            fd_start = 0  # this variable avoids an ndarray append

        fd_starts, _fds_n_rows_read = _h5_read_array(
            h5d_cl,
            fname,
            f"{oname}/cumulative_length",
            start_row=start_row,
            n_rows=n_rows,
            idx=idx2,
            use_h5idx=use_h5idx,
            obj_buf=None,
        )
        fd_starts = fd_starts.nda  # we just need the nda
        if fd_start is None:
            fd_start = fd_starts[0]

        # compute the length that flattened_data will have after the
        # fancy-indexed read
        fd_n_rows = np.sum(this_cumulen_nda[-len(fd_starts) :] - fd_starts)
        if fd_start == 0:
            fd_n_rows += this_cumulen_nda[0]

        # now make fd_idx
        fd_idx = np.empty(fd_n_rows, dtype="int32")
        fd_idx = _make_fd_idx(fd_starts, this_cumulen_nda, fd_idx)

        # Now clean up this_cumulen_nda, to be ready
        # to match the in-memory version of flattened_data. Note: these
        # operations on the view change the original array because they are
        # numpy arrays, not lists.
        this_cumulen_nda[-len(fd_starts) :] -= fd_starts
        np.cumsum(this_cumulen_nda, out=this_cumulen_nda)

    else:
        fd_idx = None

        # determine the start_row and n_rows for the flattened_data readout
        fd_start = 0
        if start_row > 0 and n_rows_read > 0:
            # need to read out the cumulen sample -before- the first sample
            # read above in order to get the starting row of the first
            # vector to read out in flattened_data
            fspace = h5d_cl.get_space()
            fspace.select_elements([[start_row - 1]])
            mspace = h5py.h5s.create(h5py.h5s.SCALAR)
            fd_start = np.empty((), h5d_cl.dtype)
            h5d_cl.read(mspace, fspace, fd_start)

            # check limits for values that will be used subsequently
            if this_cumulen_nda[-1] < fd_start:
                log.debug(
                    f"this_cumulen_nda[-1] = {this_cumulen_nda[-1]}, "
                    f"fd_start = {fd_start}, "
                    f"start_row = {start_row}, "
                    f"n_rows_read = {n_rows_read}"
                )
                msg = (
                    f"cumulative_length non-increasing between entries "
                    f"{start_row} and {start_row + n_rows_read}"
                )
                raise LH5DecodeError(msg, fname, oname)

        # Now done with this_cumulen_nda, so we can clean it up to be ready
        # to match the in-memory version of flattened_data. Note: these
        # operations on the view change the original array because they are
        # numpy arrays, not lists.
        #
        # First we need to subtract off the in-file offset for the start of
        # read for flattened_data
        this_cumulen_nda -= fd_start

        # determine the number of rows for the flattened_data readout
        fd_n_rows = this_cumulen_nda[-1] if n_rows_read > 0 else 0

    h5d_cl.close()

    # If we started with a partially-filled buffer, add the
    # appropriate offset for the start of the in-memory flattened
    # data for this read.
    fd_buf_start = np.uint32(0)
    if obj_buf_start > 0:
        fd_buf_start = cumulative_length.nda[obj_buf_start - 1]
        this_cumulen_nda += fd_buf_start

    # Now prepare the object buffer if necessary
    fd_buf = None
    if obj_buf is not None:
        fd_buf = obj_buf.flattened_data
        # grow fd_buf if necessary to hold the data
        fdb_size = fd_buf_start + fd_n_rows
        if len(fd_buf) < fdb_size:
            fd_buf.resize(fdb_size)

    # now read
    h5o = h5py.h5o.open(h5g, b"flattened_data")
    h5a_dtype = h5py.h5a.open(h5o, b"datatype")
    val = np.empty((), "O")
    h5a_dtype.read(val)
    lgdotype = dtypeutils.datatype(val.item().decode())
    if lgdotype is Array:
        _func = _h5_read_array
    elif lgdotype is VectorOfVectors:
        _func = _h5_read_vector_of_vectors
    else:
        msg = "type {lgdotype.__name__} is not supported"
        raise LH5DecodeError(msg, fname, f"{oname}/flattened_data")

    flattened_data, _ = _func(
        h5o,
        fname,
        f"{oname}/flattened_data",
        start_row=fd_start,
        n_rows=fd_n_rows,
        idx=fd_idx,
        use_h5idx=use_h5idx,
        obj_buf=fd_buf,
        obj_buf_start=fd_buf_start,
    )
    h5o.close()

    if obj_buf is not None:
        # if the buffer is partially filled, cumulative_length will be invalid
        # (i.e. non monotonically increasing). Let's fix that but filling the
        # rest of the array with the length of flattened_data
        if n_rows_read > 0:
            end = obj_buf_start + n_rows_read
            obj_buf.cumulative_length.nda[end:] = obj_buf.cumulative_length.nda[end - 1]

        return obj_buf, n_rows_read

    return (
        VectorOfVectors(
            flattened_data=flattened_data,
            cumulative_length=cumulative_length,
            attrs=read_attrs(h5g, fname, oname),
        ),
        n_rows_read,
    )


@numba.njit(parallel=False, fastmath=True)
def _make_fd_idx(starts, stops, idx):
    k = 0
    if len(starts) < len(stops):
        for i in range(stops[0]):
            idx[k] = i
            k += 1
        stops = stops[1:]
    for j in range(len(starts)):
        for i in range(starts[j], stops[j]):
            idx[k] = i
            k += 1
    return idx



from __future__ import annotations

import pickle

import pytest

import lgdo


def test_datatype_name():
    struct = lgdo.Struct()
    assert struct.datatype_name() == "struct"


def test_form_datatype():
    struct = lgdo.Struct()
    assert struct.form_datatype() == "struct{}"


def test_update_datatype():
    struct = lgdo.Struct()
    struct.update_datatype()
    assert struct.attrs["datatype"] == "struct{}"


def test_init():
    obj_dict = {"scalar1": lgdo.Scalar(value=10)}
    attrs = {"attr1": 1}
    struct = lgdo.Struct(obj_dict=obj_dict, attrs=attrs)
    assert dict(struct) == obj_dict
    assert struct.attrs == attrs | {"datatype": "struct{scalar1}"}

    with pytest.raises(ValueError):
        lgdo.Struct(obj_dict={"scalar1": 1})

    with pytest.raises(ValueError):
        lgdo.Struct(obj_dict={"scalar1": lgdo.Scalar(value=10), "thing": int})


def test_init_nested():
    obj_dict = {
        "scalar1": lgdo.Scalar(10),
        "struct1": {"field1": lgdo.Scalar(11), "field2": lgdo.Array([1, 2, 3, 4])},
    }
    struct = lgdo.Struct(obj_dict)
    assert isinstance(struct.struct1, lgdo.Struct)
    assert isinstance(struct.struct1.field1, lgdo.Scalar)
    assert struct.struct1.field1.value == 11
    assert isinstance(struct.struct1.field2, lgdo.Array)
    assert struct.struct1.field2 == lgdo.Array([1, 2, 3, 4])


def test_add_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))

    assert struct.attrs["datatype"] == "struct{scalar1}"
    assert struct["scalar1"].__class__.__name__ == "Scalar"

    struct.add_field("array1", lgdo.Array(shape=(700, 21), dtype="f", fill_val=2))
    assert struct.attrs["datatype"] == "struct{array1,scalar1}"

    struct["array2"] = lgdo.Array(shape=(700, 21), dtype="f", fill_val=2)
    assert struct.attrs["datatype"] == "struct{array1,array2,scalar1}"


def test_getattr():
    struct = lgdo.Struct()
    struct["scalar1"] = lgdo.Scalar(value=10)
    assert struct.scalar1.value == 10


def test_remove_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))
    struct.add_field("array1", lgdo.Array(shape=(10), fill_val=0))

    struct.remove_field("scalar1")
    assert list(struct.keys()) == ["array1"]

    struct.remove_field("array1", delete=True)
    assert list(struct.keys()) == []


def test_nested_access():
    struct = lgdo.Struct()
    struct.add_field("struct1/scalar1", lgdo.Scalar(value=10))
    struct.add_field("struct1/struct2/scalar2", lgdo.Scalar(value=20))
    struct.add_field("struct1/scalar3", lgdo.Scalar(value=30))

    assert "struct1" in struct
    assert "struct1/struct2" in struct
    assert "struct1/struct2/scalar2" in struct
    assert "struct1/scalar2" not in struct

    assert set(struct.keys()) == {"struct1"}
    assert set(struct["struct1"].keys()) == {"scalar1", "struct2", "scalar3"}
    assert set(struct["struct1/struct2"].keys()) == {"scalar2"}
    assert struct["struct1/scalar1"] == lgdo.Scalar(value=10)
    assert struct["struct1/struct2/scalar2"] == lgdo.Scalar(value=20)
    assert struct["struct1/scalar3"] == lgdo.Scalar(value=30)

    struct.remove_field("struct1/scalar1")
    struct.remove_field("struct1/struct2")
    assert set(struct["struct1"].keys()) == {"scalar3"}


def test_update():
    st_final = lgdo.Struct(
        {
            "a": {
                "b": lgdo.Scalar(1),
                "c": lgdo.Scalar(3),
                "d": lgdo.Scalar(5),
            }
        }
    )

    st = lgdo.Struct({"a": {"b": lgdo.Scalar(1), "c": lgdo.Scalar(2)}})
    st.update(
        lgdo.Struct(
            {
                "a": {
                    "c": lgdo.Scalar(3),
                    "d": lgdo.Scalar(5),
                }
            }
        )
    )
    assert st == st_final

    st = lgdo.Struct({"a": {"b": lgdo.Scalar(1), "c": lgdo.Scalar(2)}})
    st.update(
        {
            "a/c": lgdo.Scalar(3),
            "a/d": lgdo.Scalar(5),
        }
    )
    assert st == st_final

    st = lgdo.Struct({"a": {"b": lgdo.Scalar(1), "c": lgdo.Scalar(2)}})
    st.update(
        a={
            "c": lgdo.Scalar(3),
            "d": lgdo.Scalar(5),
        }
    )
    assert st == st_final

    st = lgdo.Struct({"a": {"b": lgdo.Scalar(1), "c": lgdo.Scalar(2)}})
    st.update(
        [
            ("a/c", lgdo.Scalar(3)),
            ("a/d", lgdo.Scalar(5)),
        ]
    )
    assert st == st_final


def test_pickle():
    obj_dict = {"scalar1": lgdo.Scalar(value=10)}
    attrs = {"attr1": 1}
    struct = lgdo.Struct(obj_dict=obj_dict, attrs=attrs)

    ex = pickle.loads(pickle.dumps(struct))
    assert isinstance(ex, lgdo.Struct)
    assert ex.attrs["attr1"] == 1
    assert ex.attrs["datatype"] == struct.attrs["datatype"]
    assert ex["scalar1"].value == 10



"""
This module implements routines from reading and writing LEGEND Data Objects in
HDF5 files.
"""

from __future__ import annotations

import logging
import sys
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from inspect import signature
from pathlib import Path
from typing import Any

import h5py
from numpy.typing import ArrayLike

from .. import types
from . import _serializers, utils
from .core import read
from .exceptions import LH5DecodeError

log = logging.getLogger(__name__)


class LH5Store:
    """
    Class to represent a store of LEGEND HDF5 files. The two main methods
    implemented by the class are :meth:`read` and :meth:`write`.

    Examples
    --------
    >>> from lgdo import LH5Store
    >>> store = LH5Store()
    >>> obj, _ = store.read("/geds/waveform", "file.lh5")
    >>> type(obj)
    lgdo.waveformtable.WaveformTable
    """

    def __init__(
        self,
        base_path: str | Path = "",
        keep_open: bool = False,
        locking: bool = False,
        default_mode: str = "r",
    ) -> None:
        """
        Parameters
        ----------
        base_path
            directory path to prepend to LH5 files.
        keep_open
            whether to keep files open by storing the :mod:`h5py` objects as
            class attributes. If ``keep_open`` is an ``int``, keep only the
            ``n`` most recently opened files; if ``True``, no limit
        locking
            whether to lock files when reading
        default_mode
            default mode in which to open files with this ``LH5Store``. See
            :class:`h5py.File` documentation. If default_mode is ``"r"``, use
            ``"a"`` when calling `LH5Store.write`.
        """
        base_path = str(Path(base_path)) if base_path != "" else ""
        self.base_path = "" if base_path == "" else utils.expand_path(base_path)
        self.keep_open = keep_open
        self.locking = locking
        self.files = OrderedDict()

        if default_mode == "read":
            default_mode = "r"
        if default_mode == "write_safe":
            default_mode = "w"
        if default_mode == "append":
            default_mode = "a"
        if default_mode == "overwrite":
            default_mode = "o"
        if default_mode == "overwrite_file":
            default_mode = "of"
        if default_mode == "append_column":
            default_mode = "ac"
        if default_mode not in ["r", "w", "a", "o", "of", "ac"]:
            msg = f"unknown wo_mode '{default_mode}'"
            raise ValueError(msg)
        self.default_mode = default_mode

    def gimme_file(
        self,
        lh5_file: str | Path | h5py.File,
        mode: str = None,
        page_buffer: int = 0,
        **file_kwargs,
    ) -> h5py.File:
        """Returns a :mod:`h5py` file object from the store or creates a new one.

        Parameters
        ----------
        lh5_file
            LH5 file name.
        mode
            mode in which to open file. See :class:`h5py.File` documentation. If
            ``None``, use default provided at construction
        page_buffer
            enable paged aggregation with a buffer of this size in bytes
            Only used when creating a new file. Useful when writing a file
            with a large number of small datasets. This is a short-hand for
            ``(fs_stragety="page", fs_pagesize=[page_buffer])``
        file_kwargs
            Keyword arguments for :class:`h5py.File`
        """
        if isinstance(lh5_file, h5py.File):
            return lh5_file

        lh5_file = str(Path(lh5_file))

        if mode is None:
            mode = self.default_mode

        if mode == "r":
            lh5_file = utils.expand_path(lh5_file, base_path=self.base_path)
            file_kwargs["locking"] = self.locking

        if lh5_file in self.files:
            self.files.move_to_end(lh5_file)
            return self.files[lh5_file]

        if self.base_path != "":
            full_path = Path(self.base_path) / lh5_file
        else:
            full_path = Path(lh5_file)

        file_exists = full_path.exists()
        if mode != "r":
            directory = full_path.parent
            if directory != "" and not full_path.parent.exists():
                log.debug(f"making path {directory}")
                directory.mkdir(parents=True, exist_ok=True)

        if mode == "r" and not file_exists:
            msg = f"file {full_path} not found"
            raise FileNotFoundError(msg)
        if not file_exists:
            mode = "w"

        if mode != "r" and file_exists:
            log.debug(f"opening existing file {full_path} in mode '{mode}'")

        if mode == "w":
            file_kwargs.update(
                {
                    "fs_strategy": "page",
                    "fs_page_size": page_buffer,
                }
            )
        try:
            h5f = h5py.File(full_path, mode, **file_kwargs)
        except (OSError, FileExistsError) as oe:
            raise LH5DecodeError(oe, full_path) from oe

        if self.keep_open:
            if isinstance(self.keep_open, int) and len(self.files) >= self.keep_open:
                self.files.popitem(last=False)
            self.files[lh5_file] = h5f

        return h5f

    def gimme_group(
        self,
        group: str | h5py.Group,
        base_group: h5py.Group,
        grp_attrs: dict[str, Any] | None = None,
        overwrite: bool = False,
    ) -> h5py.Group:
        """
        Returns an existing :class:`h5py` group from a base group or creates a new one.

        See Also
        --------
        .lh5.utils.get_h5_group
        """
        return utils.get_h5_group(group, base_group, grp_attrs, overwrite)

    def get_buffer(
        self,
        name: str,
        lh5_file: str | Path | h5py.File | Sequence[str | Path | h5py.File],
        size: int | None = None,
        field_mask: Mapping[str, bool] | Sequence[str] | None = None,
    ) -> types.LGDO:
        """Returns an LH5 object appropriate for use as a pre-allocated buffer
        in a read loop. Sets size to `size` if object has a size.
        """
        obj = self.read(name, lh5_file, n_rows=0, field_mask=field_mask)
        if hasattr(obj, "resize") and size is not None:
            obj.resize(new_size=size)
        return obj

    def read(
        self,
        name: str,
        lh5_file: str | Path | h5py.File | Sequence[str | Path | h5py.File],
        start_row: int = 0,
        n_rows: int = sys.maxsize,
        idx: ArrayLike = None,
        use_h5idx: bool = False,
        field_mask: Mapping[str, bool] | Sequence[str] | None = None,
        obj_buf: types.LGDO = None,
        obj_buf_start: int = 0,
        decompress: bool = True,
        **file_kwargs,
    ) -> tuple[types.LGDO, int]:
        """Read LH5 object data from a file in the store.

        See Also
        --------
        .lh5.core.read
        """
        # grab files from store
        if isinstance(lh5_file, (str, Path, h5py.File)):
            h5f = self.gimme_file(lh5_file, **file_kwargs)
        else:
            h5f = [self.gimme_file(f, **file_kwargs) for f in lh5_file]
        return read(
            name,
            h5f,
            start_row,
            n_rows,
            idx,
            use_h5idx,
            field_mask,
            obj_buf,
            obj_buf_start,
            decompress,
        )

    def write(
        self,
        obj: types.LGDO,
        name: str,
        lh5_file: str | Path | h5py.File,
        group: str | h5py.Group = "/",
        start_row: int = 0,
        n_rows: int | None = None,
        wo_mode: str = None,
        write_start: int = 0,
        page_buffer: int = 0,
        **h5py_kwargs,
    ) -> None:
        """Write an LGDO into an LH5 file.

        See Also
        --------
        .lh5.core.write
        """
        if wo_mode is None and self.default_mode in ["r", "read"]:
            wo_mode = "a"
        if wo_mode == "write_safe":
            wo_mode = "w"
        if wo_mode == "append":
            wo_mode = "a"
        if wo_mode == "overwrite":
            wo_mode = "o"
        if wo_mode == "overwrite_file":
            wo_mode = "of"
            write_start = 0
        if wo_mode == "append_column":
            wo_mode = "ac"
        if wo_mode not in ["w", "a", "o", "of", "ac"]:
            msg = f"unknown wo_mode '{wo_mode}'"
            raise ValueError(msg)

        # "mode" is for the h5df.File and wo_mode is for this function
        # In hdf5, 'a' is really "modify" -- in addition to appending, you can
        # change any object in the file. So we use file:append for
        # write_object:overwrite.
        mode = "w" if wo_mode == "of" else "a"

        file_kwargs = {
            k: h5py_kwargs[k]
            for k in h5py_kwargs & signature(h5py.File).parameters.keys()
        }

        return _serializers._h5_write_lgdo(
            obj,
            name,
            self.gimme_file(
                lh5_file, mode=mode, page_buffer=page_buffer, **file_kwargs
            ),
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    def read_n_rows(self, name: str, lh5_file: str | Path | h5py.File) -> int | None:
        """Look up the number of rows in an Array-like object called `name` in `lh5_file`.

        Return ``None`` if it is a :class:`.Scalar` or a :class:`.Struct`.
        """
        return utils.read_n_rows(name, self.gimme_file(lh5_file, "r"))

    def read_size_in_bytes(self, name: str, lh5_file: str | Path | h5py.File) -> int:
        """Look up the size (in B) of the object in memory. Will recursively
        crawl through all objects in a Struct or Table
        """
        return utils.read_size_in_bytes(name, self.gimme_file(lh5_file, "r"))



from __future__ import annotations

import shutil
from copy import deepcopy

import awkward as ak
import numpy as np
import pandas as pd
import pytest
from hist import axis

import lgdo
from lgdo import Table, lh5


@pytest.fixture(scope="module")
def lgnd_file(lgnd_test_data):
    return lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5")


def test_basics(lgnd_file):
    lh5_it = lh5.LH5Iterator(
        lgnd_file,
        "/geds/raw",
        entry_list=range(100),
        field_mask=["baseline"],
        buffer_len=5,
    )

    lh5_obj = lh5_it.read(4)
    assert len(lh5_obj) == 5
    assert isinstance(lh5_obj, lgdo.Table)
    assert list(lh5_obj.keys()) == ["baseline"]
    assert (
        lh5_obj["baseline"].nda == np.array([14353, 14254, 14525, 11656, 13576])
    ).all()

    for lh5_obj in lh5_it:
        entry = lh5_it.current_i_entry
        assert len(lh5_obj) == 5
        assert entry % 5 == 0
        assert all(lh5_it.current_local_entries == np.arange(entry, entry + 5))
        assert all(lh5_it.current_global_entries == np.arange(entry, entry + 5))
        assert all(lh5_it.current_files == [lgnd_file] * 5)
        assert all(lh5_it.current_groups == ["/geds/raw"] * 5)


def test_errors(lgnd_file):
    with pytest.raises(RuntimeError):
        lh5.LH5Iterator("non-existent-file.lh5", "random-group")

    with pytest.raises(ValueError):
        lh5.LH5Iterator(1, 2)

    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            lgnd_file,
            "/geds/raw",
            entry_list=range(100),
            entry_mask=np.ones(100, "bool"),
        )


def test_lgnd_waveform_table_fancy_idx(lgnd_file):
    lh5_it = lh5.LH5Iterator(
        lgnd_file,
        "geds/raw/waveform",
        entry_list=[
            7,
            9,
            25,
            27,
            33,
            38,
            46,
            52,
            57,
            59,
            67,
            71,
            72,
            82,
            90,
            92,
            93,
            94,
            97,
        ],
        buffer_len=5,
    )

    lh5_obj = lh5_it.read(0)
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert len(lh5_obj) == 5


@pytest.fixture(scope="module")
def more_lgnd_files(lgnd_test_data):
    return [
        [
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
            ),
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_raw.lh5"
            ),
        ],
        [
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_dsp.lh5"
            ),
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_dsp.lh5"
            ),
        ],
        [
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/hit/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_hit.lh5"
            ),
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/hit/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_hit.lh5"
            ),
        ],
    ]


def test_friend(more_lgnd_files):
    lh5_raw_it = lh5.LH5Iterator(
        more_lgnd_files[0],
        "ch1084803/raw",
        field_mask=["waveform", "baseline"],
        buffer_len=5,
    )
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        "ch1084803/hit",
        field_mask=["is_valid_0vbb"],
        buffer_len=5,
        friend=lh5_raw_it,
    )

    lh5_obj = lh5_it.read(0)

    assert len(lh5_obj) == 5
    assert isinstance(lh5_obj, lgdo.Table)
    assert set(lh5_obj.keys()) == {"waveform", "baseline", "is_valid_0vbb"}


def test_friend_conflict(more_lgnd_files):
    lh5_raw_it = lh5.LH5Iterator(
        more_lgnd_files[0],
        "ch1084803/raw",
        field_mask=["waveform", "baseline"],
        buffer_len=5,
    )

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        "ch1084803/dsp",
        field_mask=["baseline", "wf_max"],
        buffer_len=5,
        friend=lh5_raw_it,
    )
    lh5_obj = lh5_it.read(0)
    assert set(lh5_obj.keys()) == {"waveform", "baseline", "wf_max"}
    assert lh5_obj["waveform"] == lh5.read(
        "ch1084803/raw/waveform", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline"] == lh5.read(
        "ch1084803/dsp/baseline", more_lgnd_files[1], n_rows=5
    )
    assert lh5_obj["wf_max"] == lh5.read(
        "ch1084803/dsp/wf_max", more_lgnd_files[1], n_rows=5
    )

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        "ch1084803/dsp",
        field_mask=["baseline", "wf_max"],
        buffer_len=5,
        friend=lh5_raw_it,
        friend_prefix="raw_",
    )
    lh5_obj = lh5_it.read(0)
    assert set(lh5_obj.keys()) == {"raw_waveform", "raw_baseline", "baseline", "wf_max"}
    assert lh5_obj["raw_waveform"] == lh5.read(
        "ch1084803/raw/waveform", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["raw_baseline"] == lh5.read(
        "ch1084803/raw/baseline", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline"] == lh5.read(
        "ch1084803/dsp/baseline", more_lgnd_files[1], n_rows=5
    )
    assert lh5_obj["wf_max"] == lh5.read(
        "ch1084803/dsp/wf_max", more_lgnd_files[1], n_rows=5
    )

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        "ch1084803/dsp",
        field_mask=["baseline", "wf_max"],
        buffer_len=5,
        friend=lh5_raw_it,
        friend_suffix="_raw",
    )
    lh5_obj = lh5_it.read(0)
    assert set(lh5_obj.keys()) == {"waveform_raw", "baseline_raw", "baseline", "wf_max"}
    assert lh5_obj["waveform_raw"] == lh5.read(
        "ch1084803/raw/waveform", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline_raw"] == lh5.read(
        "ch1084803/raw/baseline", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline"] == lh5.read(
        "ch1084803/dsp/baseline", more_lgnd_files[1], n_rows=5
    )
    assert lh5_obj["wf_max"] == lh5.read(
        "ch1084803/dsp/wf_max", more_lgnd_files[1], n_rows=5
    )

    lh5_it.reset_field_mask(["waveform_raw", "baseline", "wf_max"])
    lh5_obj = lh5_it.read(0)
    assert set(lh5_obj.keys()) == {"waveform_raw", "baseline", "wf_max"}
    assert lh5_obj["waveform_raw"] == lh5.read(
        "ch1084803/raw/waveform", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline"] == lh5.read(
        "ch1084803/dsp/baseline", more_lgnd_files[1], n_rows=5
    )
    assert lh5_obj["wf_max"] == lh5.read(
        "ch1084803/dsp/wf_max", more_lgnd_files[1], n_rows=5
    )

    lh5_it.reset_field_mask({"baseline_raw": False, "wf_max": False})
    lh5_obj = lh5_it.read(0)
    assert {"waveform_raw", "baseline"}.issubset(set(lh5_obj.keys()))
    assert {"baseline_raw", "wf_max"}.isdisjoint(set(lh5_obj.keys()))
    assert lh5_obj["waveform_raw"] == lh5.read(
        "ch1084803/raw/waveform", more_lgnd_files[0], n_rows=5
    )
    assert lh5_obj["baseline"] == lh5.read(
        "ch1084803/dsp/baseline", more_lgnd_files[1], n_rows=5
    )


def test_iterate(more_lgnd_files):
    # iterate through all hit groups in all files; there are 10 entries in
    # each group/file
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        entry_list=[0, 1, 2, 3, 5, 8, 13, 21, 34, 55],
        buffer_len=5,
    )

    exp_loc_entries = [[0, 1, 2, 3, 5], [8, 3, 1, 4, 5]]
    exp_glob_entries = [[0, 1, 2, 3, 5], [8, 13, 21, 34, 55]]
    exp_files = [
        [more_lgnd_files[2][0]] * 5,
        [more_lgnd_files[2][0]] * 3 + [more_lgnd_files[2][1]] * 2,
    ]
    exp_groups = [
        ["ch1084803/hit"] * 5,
        [
            "ch1084803/hit",
            "ch1084804/hit",
            "ch1121600/hit",
            "ch1084803/hit",
            "ch1121600/hit",
        ],
    ]

    for lh5_out in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        entry = lh5_it.current_i_entry
        assert entry % 5 == 0
        assert len(lh5_out) == 5
        assert all(lh5_it.current_local_entries == exp_loc_entries[entry // 5])
        assert all(lh5_it.current_global_entries == exp_glob_entries[entry // 5])
        assert all(lh5_it.current_files == exp_files[entry // 5])
        assert all(lh5_it.current_groups == exp_groups[entry // 5])

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        entry_list=[[0, 1, 2, 3, 5, 8], [3], [1], [4], [], [5]],
        buffer_len=5,
    )

    for lh5_out in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        assert lh5_it.current_i_entry % 5 == 0
        assert len(lh5_out) == 5

    assert all(
        i == j
        for i, j in zip(
            lh5_it.get_global_entrylist(),
            [0, 1, 2, 3, 5, 8, 13, 21, 34, 55],
            strict=False,
        )
    )

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        [["ch1084803/hit", "ch1084804/hit"], ["ch1084803/hit", "ch1121600/hit"]],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
    )

    for lh5_out in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        assert lh5_it.current_i_entry % 5 == 0
        assert len(lh5_out) == 5

    # different number of file sets and group sets
    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            more_lgnd_files[2],
            [["ch1084803/hit", "ch1084804/hit"]],
            field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
            buffer_len=5,
        )

    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            more_lgnd_files[2],
            [["ch1084803/hit"], ["ch1084804/hit"], ["ch1084803/hit"]],
            field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
            buffer_len=5,
        )


def test_group_data(more_lgnd_files):
    # test provision of metadata about groups
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
        group_data={"chan": [1084803, 1084804, 1121600]},
    )

    exp_chan = [
        [1084803] * 5,
        [1084803] * 5,
        [1084804] * 5,
        [1084804] * 5,
        [1121600] * 5,
        [1121600] * 5,
        [1084803] * 5,
        [1084803] * 5,
        [1084804] * 5,
        [1084804] * 5,
        [1121600] * 5,
        [1121600] * 5,
    ]
    for tb, ec in zip(lh5_it, exp_chan, strict=False):
        assert set(tb.keys()) == {
            "is_valid_0vbb",
            "timestamp",
            "zacEmax_ctc_cal",
            "chan",
        }
        assert all(tb.chan.nda == ec)

    # group_data must be same shape as field_mask
    with pytest.raises(ValueError):
        lh5_it = lh5.LH5Iterator(
            more_lgnd_files[2],
            ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
            field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
            buffer_len=5,
            group_data={"chan": [1084803, 1084804, 1121600, 1234567]},
        )

    # group_data provided using pandas dataframe
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
        group_data=pd.DataFrame({"chan": [1084803, 1084804, 1121600]}),
    )
    for tb, ec in zip(lh5_it, exp_chan, strict=False):
        assert set(tb.keys()) == {
            "is_valid_0vbb",
            "timestamp",
            "zacEmax_ctc_cal",
            "chan",
        }
        assert all(tb.chan.nda == ec)

    # group_data provided using awkward array as array of records
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
        group_data=pd.DataFrame(
            [{"chan": 1084803}, {"chan": 1084804}, {"chan": 1121600}]
        ),
    )
    for tb, ec in zip(lh5_it, exp_chan, strict=False):
        assert set(tb.keys()) == {
            "is_valid_0vbb",
            "timestamp",
            "zacEmax_ctc_cal",
            "chan",
        }
        assert all(tb.chan.nda == ec)


def test_range(lgnd_file):
    lh5_it = lh5.LH5Iterator(
        lgnd_file,
        "/geds/raw",
        field_mask=["baseline"],
        buffer_len=5,
        i_start=7,
        n_entries=13,
    )

    # Test error when n_entries > buffer_len
    with pytest.raises(ValueError):
        lh5_obj = lh5_it.read(4, n_entries=7)

    lh5_obj = lh5_it.read(4, n_entries=3)
    assert len(lh5_obj) == 3
    assert isinstance(lh5_obj, lgdo.Table)
    assert list(lh5_obj.keys()) == ["baseline"]
    assert (lh5_obj["baseline"].nda == np.array([14353, 14254, 14525])).all()

    exp_i_entries = [7, 12, 17]
    exp_lens = [5, 5, 3]
    for lh5_obj, exp_i, exp_len in zip(lh5_it, exp_i_entries, exp_lens, strict=False):
        entry = lh5_it.current_i_entry
        assert len(lh5_obj) == exp_len
        assert entry == exp_i
        assert all(lh5_it.current_local_entries == np.arange(entry, entry + exp_len))
        assert all(lh5_it.current_global_entries == np.arange(entry, entry + exp_len))
        assert all(lh5_it.current_files == [lgnd_file] * exp_len)
        assert all(lh5_it.current_groups == ["/geds/raw"] * exp_len)


# function used in next test...
def return_tb(tb, _):
    return tb


def test_map(more_lgnd_files):
    # iterate through all hit groups in all files; there are 10 entries in
    # each group/file
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
    )

    # test that output is same with map as looping over iterator
    exp_out = []
    for tb_out in lh5_it:
        exp_out += [deepcopy(tb_out)]

    map_out = lh5_it.map(return_tb)

    assert all(
        tb_out == tb_map for tb_out, tb_map in zip(exp_out, map_out, strict=False)
    )

    # test use of append as aggregator
    tb_exp = exp_out[0]
    for tb in exp_out[1:]:
        tb_exp.append(tb)

    tb_map = lh5_it.map(return_tb, aggregate=Table.append)
    assert tb_exp == tb_map

    # test use of aggregate and init to perform reduction
    sum_exp = np.sum(tb_exp["zacEmax_ctc_cal"].nda)

    def sumE(tb, _):
        return np.sum(tb["zacEmax_ctc_cal"].nda)

    sum_map = lh5_it.map(sumE, aggregate=np.add, init=-10.0)
    assert sum_exp - 10.0 == sum_map

    # test multiprocessing; note this only works as done here if
    # buffer_len evenly divides iterator length for each file!
    map_mp = lh5_it.map(return_tb, processes=2)

    assert all(tb == tb_mp for tb, tb_mp in zip(lh5_it, map_mp, strict=False))


def query_lgdo(tb, _):
    mask = tb["zacEmax_ctc_cal"].nda > 200
    for field in tb:
        tb[field].nda = tb[field].nda[mask]
    tb.resize(np.sum(mask))
    return tb


def query_pd(tb, _):
    mask = tb["zacEmax_ctc_cal"].nda > 200
    return tb.view_as("pd").loc[mask]


def query_ak(tb, _):
    mask = tb["zacEmax_ctc_cal"].nda > 200
    return tb.view_as("ak")[mask]


def query_np(tb, _):
    mask = tb["zacEmax_ctc_cal"].nda > 200
    return tb["zacEmax_ctc_cal"].nda[mask]


def test_query(more_lgnd_files):
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
    )

    # get a table with all entries and then apply our filter
    tb_exp = None
    for tb in lh5_it:
        if tb_exp is None:
            tb_exp = deepcopy(tb)
        else:
            tb_exp.append(tb)
    tb_exp = query_lgdo(tb_exp, None)

    # filter returning Table
    tb_lgdo = lh5_it.query(query_lgdo)
    assert tb_lgdo == tb_exp
    tb_lgdo_mp = lh5_it.query(query_lgdo, processes=3)
    assert tb_lgdo_mp == tb_exp

    pd_out = lh5_it.query(query_pd)
    assert pd_out.equals(tb_exp.view_as("pd"))
    pd_out_mp = lh5_it.query(query_pd, processes=3)
    assert pd_out_mp.equals(tb_exp.view_as("pd"))

    ak_out = lh5_it.query(query_ak)
    assert ak.array_equal(ak_out, tb_exp.view_as("ak"))
    ak_out_mp = lh5_it.query(query_ak, processes=3)
    assert ak.array_equal(ak_out_mp, tb_exp.view_as("ak"))

    np_out = lh5_it.query(query_np)
    assert np.all(np_out == tb_exp["zacEmax_ctc_cal"])
    np_out_mp = lh5_it.query(query_np, processes=3)
    assert np.all(np_out_mp == tb_exp["zacEmax_ctc_cal"])

    pd_out_str = lh5_it.query("zacEmax_ctc_cal>200", library="pd")
    assert pd_out_str.equals(tb_exp.view_as("pd"))
    pd_out_str_mp = lh5_it.query("zacEmax_ctc_cal>200", processes=3, library="pd")
    assert pd_out_str_mp.equals(tb_exp.view_as("pd"))


def test_hist(more_lgnd_files):
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[2],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
    )

    # get a table with all entries and then apply our filter
    tb_exp = None
    for tb in lh5_it:
        if tb_exp is None:
            tb_exp = deepcopy(tb)
        else:
            tb_exp.append(tb)
    tb_exp = query_lgdo(tb_exp, None)

    h_exp, xedges, yedges = np.histogram2d(
        tb_exp["timestamp"], tb_exp["zacEmax_ctc_cal"]
    )
    # otherwise some entries end up in the overflow/underflow bins
    xedges[0] *= 0.99
    xedges[-1] *= 1.01
    yedges[0] *= 0.99
    yedges[-1] *= 1.01

    h_lgdo = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_lgdo,
        keys=["timestamp", "zacEmax_ctc_cal"],
    )
    assert np.all(np.array(h_lgdo) == h_exp)
    h_lgdo_mp = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_lgdo,
        keys=["timestamp", "zacEmax_ctc_cal"],
        processes=2,
    )
    assert np.all(np.array(h_lgdo_mp) == h_exp)

    h_pd = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_pd,
        keys=["timestamp", "zacEmax_ctc_cal"],
    )
    assert np.all(np.array(h_pd) == h_exp)
    h_pd_mp = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_pd,
        keys=["timestamp", "zacEmax_ctc_cal"],
        processes=2,
    )
    assert np.all(np.array(h_pd_mp) == h_exp)

    h_ak = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_ak,
        keys=["timestamp", "zacEmax_ctc_cal"],
    )
    assert np.all(np.array(h_ak) == h_exp)
    h_ak_mp = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where=query_ak,
        keys=["timestamp", "zacEmax_ctc_cal"],
        processes=2,
    )
    assert np.all(np.array(h_ak_mp) == h_exp)

    h_np = lh5_it.hist(
        [axis.Variable(yedges)],
        where=query_np,
        keys=["zacEmax_ctc_cal"],
    )
    assert np.all(np.array(h_np) == np.sum(h_exp, axis=0))
    h_np_mp = lh5_it.hist(
        [axis.Variable(yedges)],
        where=query_np,
        keys=["zacEmax_ctc_cal"],
        processes=2,
    )
    assert np.all(np.array(h_np_mp) == np.sum(h_exp, axis=0))

    h_pd_str = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where="zacEmax_ctc_cal>200",
        keys=["timestamp", "zacEmax_ctc_cal"],
    )
    assert np.all(np.array(h_pd_str) == h_exp)
    h_pd_str_mp = lh5_it.hist(
        [axis.Variable(xedges), axis.Variable(yedges)],
        where="zacEmax_ctc_cal>200",
        keys=["timestamp", "zacEmax_ctc_cal"],
        processes=2,
    )
    assert np.all(np.array(h_pd_str_mp) == h_exp)


def test_iterator_wo_mode_write(tmp_path, lh5_file):
    dst = tmp_path / "rw.lh5"
    shutil.copy(lh5_file, dst)
    it = lh5.LH5Iterator(
        dst.as_posix(), "/data/struct_full/array", h5py_open_mode="append"
    )
    store = it.lh5_st
    store.write(
        lgdo.Array(nda=np.array([0], dtype=int)),
        "dummy",
        dst.as_posix(),
        group="/data",
        wo_mode="append",
    )
    assert len(store.read("/data/dummy", dst.as_posix())) == 1
    assert len(it.read(0)) > 0



from __future__ import annotations

import pytest

from lgdo import types
from lgdo.lh5 import datatype


def test_datatype2lgdo():
    d = datatype.datatype
    assert d("real") == types.Scalar
    assert d("bool") == types.Scalar
    assert d("complex") == types.Scalar

    assert d(" real ") == types.Scalar

    with pytest.raises(RuntimeError):
        assert d("int") == types.Scalar

    assert d("array<1>{real}") == types.Array
    assert d("fixedsize_array<1>{real}") == types.FixedSizeArray
    assert d("array_of_equalsized_arrays<1,1>{real}") == types.ArrayOfEqualSizedArrays

    assert d("array<1>{array<1>{real}}") == types.VectorOfVectors
    assert d("array<1>{array<1>{array<1>{real}}}") == types.VectorOfVectors

    assert d("array<1>{encoded_array<1>{real}}") == types.VectorOfEncodedVectors
    assert (
        d("array_of_encoded_equalsized_arrays<1,1>{real}")
        == types.ArrayOfEncodedEqualSizedArrays
    )

    assert d("struct{a,b,c,d}") == types.Struct
    assert d("struct{}") == types.Struct
    assert d("table{a,b,c,d}") == types.Table
    assert d("struct{binning,weights,isdensity}") == types.Histogram


def test_utils():
    assert (
        datatype.get_nested_datatype_string("array<1>{encoded_array<1>{real}}")
        == "encoded_array<1>{real}"
    )
    assert datatype.get_nested_datatype_string("table{a,b,c,d}") == "a,b,c,d"
    assert datatype.get_nested_datatype_string("table{}") == ""
    assert (
        datatype.get_nested_datatype_string("array_of_equalsized_arrays<1,1>{real}")
        == "real"
    )

    assert datatype.get_struct_fields("table{a,b,c,d}") == ["a", "b", "c", "d"]
    assert datatype.get_struct_fields("table{}") == []



"""LEGEND Data Objects (LGDO) types."""

from __future__ import annotations

from .array import Array
from .arrayofequalsizedarrays import ArrayOfEqualSizedArrays
from .encoded import ArrayOfEncodedEqualSizedArrays, VectorOfEncodedVectors
from .fixedsizearray import FixedSizeArray
from .histogram import Histogram
from .lgdo import LGDO, LGDOCollection
from .scalar import Scalar
from .struct import Struct
from .table import Table
from .vectorofvectors import VectorOfVectors
from .waveformtable import WaveformTable

__all__ = [
    "LGDO",
    "Array",
    "ArrayOfEncodedEqualSizedArrays",
    "ArrayOfEqualSizedArrays",
    "FixedSizeArray",
    "Histogram",
    "LGDOCollection",
    "Scalar",
    "Struct",
    "Table",
    "VectorOfEncodedVectors",
    "VectorOfVectors",
    "WaveformTable",
]

import numpy as np

np.set_printoptions(threshold=10)



from __future__ import annotations

import logging
import sys
from bisect import bisect_left

import h5py
import hdf5plugin  # noqa: F401, used for compression plugins
import numpy as np

from ....types import Array
from ... import datatype
from ...exceptions import LH5DecodeError
from .utils import read_attrs

log = logging.getLogger(__name__)


def _h5_read_ndarray(
    h5d,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    obj_buf=None,
    obj_buf_start=0,
):
    if obj_buf is not None and not isinstance(obj_buf, Array):
        msg = "object buffer is not an Array"
        raise LH5DecodeError(msg, fname, oname)

    # compute the number of rows to read
    # we culled idx above for start_row and n_rows, now we have to apply
    # the constraint of the length of the dataset
    try:
        fspace = h5d.get_space()
        ds_n_rows = fspace.shape[0]
    except AttributeError as e:
        msg = "does not seem to be an HDF5 dataset"
        raise LH5DecodeError(msg, fname, oname) from e

    if idx is not None:
        if len(idx) > 0 and idx[-1] >= ds_n_rows:
            log.warning("idx indexed past the end of the array in the file. Culling...")
            n_rows_to_read = bisect_left(idx, ds_n_rows)
            idx = idx[:n_rows_to_read]
            if len(idx) == 0:
                log.warning("idx empty after culling.")
        n_rows_to_read = len(idx)
    else:
        n_rows_to_read = ds_n_rows - start_row
    n_rows_to_read = min(n_rows_to_read, n_rows)

    if idx is None:
        fspace.select_hyperslab(
            (start_row,) + (0,) * (h5d.rank - 1),
            (1,) * h5d.rank,
            None,
            (n_rows_to_read, *fspace.shape[1:]),
        )
    elif use_h5idx:
        # Note that h5s will automatically merge adjacent elements into a range
        fspace.select_none()
        for i in idx:
            fspace.select_hyperslab(
                (i,) + (0,) * (h5d.rank - 1),
                (1,) * h5d.rank,
                None,
                (1, *fspace.shape[1:]),
                h5py.h5s.SELECT_OR,
            )

    # Now read the array
    if obj_buf is not None and n_rows_to_read > 0:
        buf_size = obj_buf_start + n_rows_to_read
        if len(obj_buf) < buf_size:
            obj_buf.resize(buf_size)
        dest_sel = np.s_[obj_buf_start:buf_size]

        if idx is None or use_h5idx:
            mspace = h5py.h5s.create_simple(obj_buf.nda.shape)
            mspace.select_hyperslab(
                (obj_buf_start,) + (0,) * (h5d.rank - 1),
                (1,) * h5d.rank,
                None,
                (n_rows_to_read, *fspace.shape[1:]),
            )
            h5d.read(mspace, fspace, obj_buf.nda)
        else:
            tmp = np.empty(fspace.shape, h5d.dtype)
            h5d.read(fspace, fspace, tmp)
            obj_buf.nda[dest_sel, ...] = tmp[idx, ...]
        nda = obj_buf.nda
    elif n_rows == 0:
        tmp_shape = (0, *h5d.shape[1:])
        nda = np.empty(tmp_shape, h5d.dtype)
    else:
        mspace = h5py.h5s.create_simple((n_rows_to_read, *fspace.shape[1:]))
        nda = np.empty(mspace.shape, h5d.dtype)
        if idx is None or use_h5idx:
            h5d.read(mspace, fspace, nda)
        else:
            tmp = np.empty(fspace.shape, h5d.dtype)
            h5d.read(fspace, fspace, tmp)
            nda[:, ...] = tmp[idx, ...]

    # Finally, set attributes and return objects
    attrs = read_attrs(h5d, fname, oname)

    # special handling for bools
    # (c and Julia store as uint8 so cast to bool)
    if datatype.get_nested_datatype_string(attrs["datatype"]) == "bool":
        nda = nda.astype(np.bool_, copy=False)

    return (nda, attrs, n_rows_to_read)



"""
Implements a LEGEND Data Object representing an n-dimensional array and
corresponding utilities.
"""

from __future__ import annotations

import logging
from collections.abc import Collection, Iterator
from typing import Any

import awkward as ak
import awkward_pandas as akpd
import numpy as np
import pandas as pd
import pint_pandas  # noqa: F401

from .. import utils
from ..units import default_units_registry as u
from .lgdo import LGDOCollection

log = logging.getLogger(__name__)


class Array(LGDOCollection):
    r"""Holds an :class:`numpy.ndarray` and attributes.

    :class:`Array` (and the other various array types) holds an `nda` instead
    of deriving from :class:`numpy.ndarray` for the following reasons:

    - It keeps management of the `nda` totally under the control of the user. The
      user can point it to another object's buffer, grab the `nda` and toss the
      :class:`Array`, etc.
    - It allows the management code to send just the `nda`'s the central routines
      for data manpulation. Keeping LGDO's out of that code allows for more
      standard, reusable, and (we expect) performant Python.
    - It allows the first axis of the `nda` to be treated as "special" for storage
      in :class:`.Table`\ s.
    """

    def __init__(
        self,
        nda: np.ndarray | ak.Array | None = None,
        shape: tuple[int, ...] = (),
        dtype: np.dtype | None = None,
        fill_val: float | int | None = None,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        nda
            An :class:`numpy.ndarray` or :class:`ak.Array` to be used for this
            object's internal array. If the Awkward array carries a ``units``
            parameter, it will forwarded as LGDO attribute.
        shape
            A numpy-format shape specification for shape of the internal
            ndarray. Required if `nda` is ``None``, otherwise unused.
        dtype
            Specifies the type of the data in the array. Required if `nda` is
            ``None``, otherwise unused.
        fill_val
            If ``None``, memory is allocated without initialization. Otherwise,
            the array is allocated with all elements set to the corresponding
            fill value. If `nda` is not ``None``, this parameter is ignored.
        attrs
            A set of user attributes to be carried along with this LGDO. These
            attributes have always precedence over all the others (e.g. those
            carried by `nda`).

        Warning
        -------
        This constructor has partial units support. It supports fishing `units`
        from Awkward Array parameters but not (yet) from e.g. NumPy+Pint
        arrays. In any case, the user can always attach units later by
        modifying the dictionary held by :attr:`attrs`.

        Note
        ----
        The array is used directly, not copied. If not supplied, internal
        memory is newly allocated based on the shape and dtype arguments.
        """
        if nda is None:
            if fill_val is None:
                nda = np.empty(shape, dtype=dtype)
            elif fill_val == 0:
                nda = np.zeros(shape, dtype=dtype)
            else:
                nda = np.full(shape, fill_val, dtype=dtype)

        elif isinstance(nda, ak.Array):
            # units: we don't just forward all ak parameters, there might be
            # some weird thing in there
            units = ak.parameters(nda).get("units", None)
            if units is not None:
                if attrs is None:
                    attrs = {}

                # give precedence to the user units
                attrs = {"units": units} | attrs

            if nda.type.content.parameters.get("__array__") == "string":
                # Variable length strings aren't quite up to snuff yet, so pad the
                # fixed-width string length in case we want to update the array
                # TODO: numpy 2.2.5 fixes this; but it required python v3.10 or higher
                s_len = np.max(ak.num(nda))
                nda = np.array(nda, dtype=f"<U{s_len * 2}")
            else:
                nda = ak.to_numpy(nda)  # this is zero-copy

        elif isinstance(nda, Array):
            nda = nda.nda

        self.nda = nda

        super().__init__(attrs)

    def datatype_name(self) -> str:
        return "array"

    def form_datatype(self) -> str:
        dt = self.datatype_name()
        nd = str(len(self.nda.shape))
        et = utils.get_element_type(self)
        return dt + "<" + nd + ">{" + et + "}"

    def __len__(self) -> int:
        return self._size

    @property
    def nda(self):
        return self._nda[: self._size, ...] if self._nda.shape != () else self._nda

    @nda.setter
    def nda(self, value):
        self._nda = value if isinstance(value, np.ndarray) else np.array(value)
        self._size = len(self._nda) if self._nda.shape != () else 0

    @property
    def dtype(self):
        return self._nda.dtype

    @property
    def shape(self):
        return (len(self), *self._nda.shape[1:])

    def reserve_capacity(self, capacity: int) -> None:
        "Set size (number of rows) of internal memory buffer"
        if capacity < len(self):
            msg = "Cannot reduce capacity below Array length"
            raise ValueError(msg)
        self._nda.resize((capacity, *self._nda.shape[1:]), refcheck=False)

    def get_capacity(self) -> int:
        "Get capacity (i.e. max size before memory must be re-allocated)"
        return len(self._nda)

    def trim_capacity(self) -> None:
        "Set capacity to be minimum needed to support Array size"
        self.reserve_capacity(np.prod(self.shape))

    def resize(self, new_size: int | Collection[int], trim=False) -> None:
        """Set size of Array in rows. Only change capacity if it must be
        increased to accommodate new rows; in this case double capacity.
        If trim is True, capacity will be set to match size. If new_size
        is an int, do not change size of inner dimensions.

        If new_size is a collection, internal memory will be re-allocated, so
        this should be done only rarely!

        Parameters
        ----------
        new_size
            number of rows after resize. If provided as a collection, also
            resize inner dimensions
        trim
            after resizing, change capacity to match size
        """

        if isinstance(new_size, Collection):
            self._size = new_size[0]
            self._nda.resize(new_size)
        else:
            self._size = new_size

            if trim and new_size != self.get_capacity:
                self.reserve_capacity(new_size)

            # If capacity is not big enough, set to next power of 2 big enough
            if new_size > self.get_capacity():
                self.reserve_capacity(int(2 ** (np.ceil(np.log2(new_size)))))

    def append(self, value: Array | np.scalar | np.ndarray) -> None:
        "Append value(s) to end of array (with copy), resizing array"
        self.insert(len(self), value)

    def insert(self, i: int, value: Array | np.scalar | np.ndarray) -> None:
        """Insert value(s) into row i (with copy). If collection of values are provided
        insert all values. Array will be resized and subsequent values will be shifted.
        """
        if i > len(self):
            msg = f"index {i} is out of bounds for array with size {len(self)}"
            raise IndexError(msg)

        if not isinstance(value, (Array, np.ndarray)):
            value = np.array(value, dtype=self.dtype)

        if value.shape == self.shape[1:]:
            self.resize(len(self) + 1)
            self[i + 1 :] = self[i:-1]
            self[i] = value
        elif value.shape[1:] == self.shape[1:]:
            self.resize(len(self) + len(value))
            self[i + len(value) :] = self[i : -len(value)]
            self[i : i + len(value)] = value
        else:
            msg = f"Could not insert value with shape {value.shape} into Array with shape {self.shape}"
            raise ValueError(msg)

    def replace(self, i: int, value: np.scalar | np.ndarray) -> None:
        "Replace value at row i"
        if i >= len(self):
            msg = f"index {i} is out of bounds for array with size {len(self)}"
            raise IndexError(msg)
        self[i] = value

    def __getitem__(self, key):
        return self.nda[key]

    def __setitem__(self, key, value):
        return self.nda.__setitem__(key, value)

    def __eq__(self, other: Array) -> bool:
        if isinstance(other, Array):
            return self.attrs == other.attrs and np.array_equal(self.nda, other.nda)

        return False

    def __hash__(self):
        return hash(self.name)

    def __iter__(self) -> Iterator:
        yield from self.nda

    def __str__(self) -> str:
        attrs = self.getattrs()
        string = str(self.nda)
        if attrs:
            string += f" with attrs={attrs}"
        return string

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + np.array2string(
                self.nda,
                prefix=self.__class__.__name__ + " ",
                formatter={
                    "int": lambda x: f"0x{x:02x}" if self.dtype == np.ubyte else str(x)
                },
            )
            + f", attrs={self.attrs!r})"
        )

    def view_as(
        self, library: str, with_units: bool = False
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        """View the Array data as a third-party format data structure.

        This is a zero-copy operation. Supported third-party formats are:

        - ``pd``: returns a :class:`pandas.Series`
        - ``np``: returns the internal `nda` attribute (:class:`numpy.ndarray`)
        - ``ak``: returns an :class:`ak.Array` initialized with `self.nda`

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.

        See Also
        --------
        .LGDO.view_as
        """
        attach_units = with_units and "units" in self.attrs

        if library == "pd":
            if attach_units:
                if self.nda.ndim == 1:
                    return pd.Series(
                        self.nda, dtype=f"pint[{self.attrs['units']}]", copy=False
                    )

                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            if self.nda.ndim == 1:
                return pd.Series(self.nda, copy=False)

            # if array is multi-dim, use awkward
            return akpd.from_awkward(self.view_as("ak"))

        if library == "np":
            if attach_units:
                # TODO: does attaching units imply a copy?
                return self.nda * u(self.attrs["units"])

            return self.nda

        if library == "ak":
            ak_arr = ak.Array(self.nda)  # NOTE: this is zero-copy!
            if attach_units:
                ak_arr = ak.with_parameter(ak_arr, "units", self.attrs["units"])

            return ak_arr

        msg = f"{library} is not a supported third-party format."
        raise ValueError(msg)



from __future__ import annotations

import h5py


class LH5DecodeError(Exception):
    def __init__(
        self, message: str, file: str | h5py.File, oname: str | None = None
    ) -> None:
        super().__init__(message)

        self.file = file.filename if isinstance(file, h5py.File) else file
        self.obj = oname

    def __str__(self) -> str:
        if self.obj is None:
            msg = f"while opening file {self.file} for decoding: "
        else:
            msg = f"while decoding object '{self.obj}' in file {self.file}: "

        return msg + super().__str__()

    def __reduce__(self) -> tuple:  # for pickling.
        return self.__class__, (*self.args, self.file, self.obj)


class LH5EncodeError(Exception):
    def __init__(
        self,
        message: str,
        file: str | h5py.File,
        group: str | h5py.Group | None = None,
        name: str | None = None,
    ) -> None:
        super().__init__(message)

        self.file = file.filename if isinstance(file, h5py.File) else file
        self.group = (
            (group.name if isinstance(file, h5py.File) else group).rstrip("/")
            if group is not None
            else None
        )
        self.name = name.lstrip("/") if name is not None else None

    def __str__(self) -> str:
        if self.name is None:
            msg = f"while opening file {self.file} for encoding: "
        else:
            msg = (
                f"while encoding object {self.group}/{self.name} to file {self.file}: "
            )
        return msg + super().__str__()

    def __reduce__(self) -> tuple:  # for pickling.
        return self.__class__, (*self.args, self.file, self.group, self.name)



from __future__ import annotations

from typing import Any


def default_hdf5_settings() -> dict[str, Any]:
    """Returns the HDF5 settings for writing data to disk to the pydataobj defaults.

    Examples
    --------
    >>> from lgdo import lh5
    >>> lh5.DEFAULT_HDF5_SETTINGS["compression"] = "lzf"
    >>> lh5.write(data, "data", "file.lh5")  # compressed with LZF
    >>> lh5.DEFAULT_HDF5_SETTINGS = lh5.default_hdf5_settings()
    >>> lh5.write(data, "data", "file.lh5", "of")  # compressed with default settings (GZIP)
    """

    return {
        "shuffle": True,
        "compression": "gzip",
    }


DEFAULT_HDF5_SETTINGS: dict[str, ...] = default_hdf5_settings()
"""Global dictionary storing the default HDF5 settings for writing data to disk.

Modify this global variable before writing data to disk with this package.

Examples
--------
>>> from lgdo import lh5
>>> lh5.DEFAULT_HDF5_SETTINGS["compression"] = "lzf"
>>> lh5.write(data, "data", "file.lh5")  # compressed with LZF
"""



from __future__ import annotations

import logging

import h5py
import numpy as np

from .... import types
from ... import settings
from ...exceptions import LH5EncodeError

log = logging.getLogger(__name__)


def _h5_write_array(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.Array)

    if n_rows is None or n_rows > obj.nda.shape[0] - start_row:
        n_rows = obj.nda.shape[0] - start_row

    nda = obj.nda[start_row : start_row + n_rows]

    # hack to store bools as uint8 for c / Julia compliance
    if nda.dtype.name == "bool":
        nda = nda.astype(np.uint8)

    # need to create dataset from ndarray the first time for speed
    # creating an empty dataset and appending to that is super slow!
    if (wo_mode != "a" and write_start == 0) or name not in group:
        # this is needed in order to have a resizable (in the first
        # axis) data set, i.e. rows can be appended later
        # NOTE: this automatically turns chunking on!
        maxshape = (None, *nda.shape[1:])
        h5py_kwargs.setdefault("maxshape", maxshape)

        if wo_mode == "o" and name in group:
            log.debug(f"overwriting {name} in {group}")
            del group[name]

        # set default compression options
        for k, v in settings.DEFAULT_HDF5_SETTINGS.items():
            h5py_kwargs.setdefault(k, v)

        # compress using the 'compression' LGDO attribute, if available
        if "compression" in obj.attrs:
            comp_algo = obj.attrs["compression"]
            if isinstance(comp_algo, dict):
                h5py_kwargs |= obj.attrs["compression"]
            else:
                h5py_kwargs["compression"] = obj.attrs["compression"]

        # and even the 'hdf5_settings' one, preferred
        if "hdf5_settings" in obj.attrs:
            h5py_kwargs |= obj.attrs["hdf5_settings"]

        # HACK: a tuple is strictly requested for the "chunks" setting, but
        # we'd like to pass a list too in some situations
        if "chunks" in h5py_kwargs and isinstance(h5py_kwargs["chunks"], list):
            h5py_kwargs["chunks"] = tuple(h5py_kwargs["chunks"])

        # create HDF5 dataset
        ds = group.create_dataset(name, data=nda, **h5py_kwargs)

        # attach HDF5 dataset attributes, but not "compression"!
        _attrs = obj.getattrs(datatype=True)
        _attrs.pop("compression", None)
        _attrs.pop("hdf5_settings", None)
        ds.attrs.update(
            {
                k: v.encode("utf-8") if isinstance(v, str) else v
                for k, v in _attrs.items()
            }
        )

        return

    # Now append or overwrite
    ds = group[name]
    if not isinstance(ds, h5py.Dataset):
        msg = (
            f"existing HDF5 object '{name}' in group '{group}'"
            " is not a dataset! Cannot overwrite or append"
        )
        raise LH5EncodeError(msg, lh5_file, group, name)

    old_len = ds.shape[0]
    if wo_mode == "a":
        write_start = old_len

    ds.resize(write_start + nda.shape[0], axis=0)
    ds[write_start:] = nda



r"""Data compression utilities.

This subpackage collects all LEGEND custom data compression (encoding) and
decompression (decoding) algorithms.

Available lossless waveform compression algorithms:

* :class:`.RadwareSigcompress`, a Python port of the C algorithm
  `radware-sigcompress` by D. Radford.
* :class:`.ULEB128ZigZagDiff` variable-length base-128 encoding of waveform
  differences.

All waveform compression algorithms inherit from the :class:`.WaveformCodec`
abstract class.

:func:`~.generic.encode` and :func:`~.generic.decode` provide a high-level
interface for encoding/decoding :class:`~.lgdo.LGDO`\ s.

>>> from lgdo import WaveformTable, compression
>>> wftbl = WaveformTable(...)
>>> enc_wft = compression.encode(wftable, RadwareSigcompress(codec_shift=-23768)
>>> compression.decode(enc_wft) # == wftbl
"""

from __future__ import annotations

from importlib import import_module

from .base import WaveformCodec
from .generic import decode, encode

# mapping from codec class to module for lazy loading
_codec_classes = {
    "RadwareSigcompress": "radware",
    "ULEB128ZigZagDiff": "varlen",
}

__all__ = list(_codec_classes)
__all__ += [
    "WaveformCodec",
    "decode",
    "encode",
]


# Lazy loader
def __getattr__(name):
    if name in _codec_classes:
        mod_name = _codec_classes[name]
        mod = import_module(f".{mod_name}", __name__)
        codec = getattr(mod, name)
        globals().update({name: codec})
        return codec
    msg = f"module {__name__} has no attribute {name}"
    raise AttributeError(msg)


def __dir__():
    return __all__ + list(globals().keys())



from __future__ import annotations

import logging

from .. import types as lgdo
from .base import WaveformCodec

log = logging.getLogger(__name__)


def encode(
    obj: lgdo.VectorOfVectors | lgdo.ArrayOfEqualsizedArrays,
    codec: WaveformCodec | str = None,
) -> lgdo.VectorOfEncodedVectors | lgdo.ArrayOfEncodedEqualSizedArrays:
    """Encode LGDOs with `codec`.

    Defines behaviors for each implemented waveform encoding algorithm.

    Parameters
    ----------
    obj
        LGDO array type.
    codec
        algorithm to be used for encoding.
    """
    log.debug(f"encoding {obj!r} with {codec}")

    from . import radware, varlen  # noqa: PLC0415

    if _is_codec(codec, radware.RadwareSigcompress):
        enc_obj = radware.encode(obj, shift=codec.codec_shift)
    elif _is_codec(codec, varlen.ULEB128ZigZagDiff):
        enc_obj = varlen.encode(obj)
    else:
        msg = f"'{codec}' not supported"
        raise ValueError(msg)

    enc_obj.attrs |= codec.asdict()

    return enc_obj


def decode(
    obj: lgdo.VectorOfEncodedVectors | lgdo.ArrayOfEncodedEqualSizedArrays,
    out_buf: lgdo.ArrayOfEqualSizedArrays = None,
) -> lgdo.VectorOfVectors | lgdo.ArrayOfEqualsizedArrays:
    """Decode encoded LGDOs.

    Defines decoding behaviors for each implemented waveform encoding
    algorithm. Expects to find the codec (and its parameters) the arrays where
    encoded with among the LGDO attributes.

    Parameters
    ----------
    obj
        LGDO array type.
    out_buf
        pre-allocated LGDO for the decoded signals. See documentation of
        wrapped encoders for limitations.
    """
    if "codec" not in obj.attrs:
        msg = (
            "object does not carry any 'codec' attribute, I don't know how to decode it"
        )
        raise RuntimeError(msg)

    codec = obj.attrs["codec"]
    log.debug(f"decoding {obj!r} with {codec}")

    from . import radware, varlen  # noqa: PLC0415

    if _is_codec(codec, radware.RadwareSigcompress):
        return radware.decode(
            obj, sig_out=out_buf, shift=int(obj.attrs.get("codec_shift", 0))
        )

    if _is_codec(codec, varlen.ULEB128ZigZagDiff):
        return varlen.decode(obj, sig_out=out_buf)

    msg = f"'{codec}' not supported"
    raise ValueError(msg)


def _is_codec(ident: WaveformCodec | str, codec) -> bool:
    if isinstance(ident, WaveformCodec):
        return isinstance(ident, codec)

    if isinstance(ident, str):
        return ident == codec().codec

    msg = "input must be WaveformCodec object or string identifier"
    raise ValueError(msg)



from __future__ import annotations

from collections import namedtuple

import numpy as np
import pytest

from lgdo import VectorOfVectors
from lgdo.types import vovutils

VovColl = namedtuple("VovColl", ["v2d", "v3d", "v4d"])


@pytest.fixture
def testvov():
    v2d = VectorOfVectors([[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]])
    v3d = VectorOfVectors([[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]])
    v4d = VectorOfVectors(
        [
            [[[1], [2]], [[3, 4], [5]]],
            [[[2, 6]], [[4, 8, 9, 7], [8, 3]]],
            [[[5, 3], [1]]],
        ]
    )

    return VovColl(v2d, v3d, v4d)


def test_nb_fill():
    # test 1d array of int
    aoa_in = np.arange(5, dtype="int32").reshape(1, 5)
    len_in = np.array([5])
    nan_val = np.array([0], dtype=aoa_in.dtype)
    flattened_array_out = np.empty(5, dtype=aoa_in.dtype)

    vovutils._nb_fill(aoa_in, len_in, nan_val, flattened_array_out)
    assert np.array_equal(
        flattened_array_out, np.array([0, 1, 2, 3, 4], dtype=aoa_in.dtype)
    )
    # test 1d array of uint
    aoa_in = np.arange(5, dtype="uint16").reshape(1, 5)
    len_in = np.array([5])
    nan_val = np.array([0], dtype=aoa_in.dtype)
    flattened_array_out = np.empty(5, dtype=aoa_in.dtype)

    vovutils._nb_fill(aoa_in, len_in, nan_val, flattened_array_out)
    assert np.array_equal(
        flattened_array_out, np.array([0, 1, 2, 3, 4], dtype=aoa_in.dtype)
    )
    # test 1d array of float
    aoa_in = np.arange(5, dtype="float32").reshape(1, 5)
    len_in = np.array([5])
    nan_val = np.array([0], dtype=aoa_in.dtype)
    flattened_array_out = np.empty(5, dtype=aoa_in.dtype)

    vovutils._nb_fill(aoa_in, len_in, nan_val, flattened_array_out)
    assert np.array_equal(
        flattened_array_out, np.array([0, 1, 2, 3, 4], dtype=aoa_in.dtype)
    )
    # test 2d array of int
    aoa_in = np.array([[1, 2, 3], [4, 5, 6]], dtype="int32")
    len_in = np.array([3, 3])
    nan_val = np.array([0], dtype=aoa_in.dtype)
    flattened_array_out = np.empty(6, dtype=aoa_in.dtype)
    vovutils._nb_fill(aoa_in, len_in, nan_val, flattened_array_out)
    assert np.array_equal(
        flattened_array_out,
        np.array([1, 2, 3, 4, 5, 6], dtype=aoa_in.dtype),
    )
    # test nan value addition
    aoa_in = np.array([[1, 2, 3], [4, 5, 6]], dtype="int32")
    len_in = np.array([4, 3])
    flattened_array_out = np.empty(7, dtype=aoa_in[0].dtype)
    vovutils._nb_fill(aoa_in, len_in, nan_val, flattened_array_out)
    assert np.array_equal(
        flattened_array_out,
        np.array([1, 2, 3, 0, 4, 5, 6], dtype=aoa_in[0].dtype),
    )


def test_build_cl_and_explodes():
    cl = np.array([3, 4], dtype=np.uint64)
    exp = np.array([0, 0, 0, 1], dtype=np.uint64)
    array = np.array([5, 7], dtype=np.uint64)
    array_exp = np.array([5, 5, 5, 7], dtype=np.uint64)
    # build_cl
    assert (vovutils.build_cl(exp, cl) == cl).all()
    assert (vovutils.build_cl(exp) == cl).all()
    assert (vovutils.build_cl([0, 0, 0, 1]) == cl).all()
    assert (vovutils.build_cl(array_exp, cl) == cl).all()
    assert (vovutils.build_cl(array_exp) == cl).all()
    assert (vovutils.build_cl([5, 5, 5, 7]) == cl).all()
    # explode_cl
    assert (vovutils.explode_cl(cl, exp) == exp).all()
    assert (vovutils.explode_cl(cl) == exp).all()
    assert (vovutils.explode_cl([3, 4]) == exp).all()
    # inverse functionality
    assert (vovutils.build_cl(vovutils.explode_cl(cl)) == cl).all()
    assert (vovutils.explode_cl(vovutils.build_cl(array_exp)) == exp).all()
    # explode
    assert (vovutils.explode(cl, array, array_exp) == array_exp).all()
    assert (vovutils.explode(cl, array) == array_exp).all()
    assert (vovutils.explode([3, 4], [5, 7]) == array_exp).all()
    assert (vovutils.explode(cl, range(len(cl))) == exp).all()
    # explode_arrays
    arrays_out = vovutils.explode_arrays(cl, [array, range(len(cl))])
    assert len(arrays_out) == 2
    assert (arrays_out[0] == array_exp).all()
    assert (arrays_out[1] == exp).all()
    arrays_out = vovutils.explode_arrays(
        cl, [array, range(len(cl))], arrays_out=arrays_out
    )
    assert len(arrays_out) == 2
    assert (arrays_out[0] == array_exp).all()
    assert (arrays_out[1] == exp).all()



from __future__ import annotations

import dbetto
import hist
import numpy as np
import pytest

import lgdo


def test_eval_dependency():
    obj = lgdo.Table(
        col_dict={
            "a": lgdo.Array([1, 2, 3, 4], attrs={"units": "s"}),
            "b": lgdo.Array([5, 6, 7, 8]),
            "c": lgdo.ArrayOfEqualSizedArrays(
                nda=[[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
            ),
            "d": lgdo.ArrayOfEqualSizedArrays(
                nda=[
                    [21, 22, 23, 24],
                    [25, 26, 27, 8],
                    [29, 210, 211, 212],
                    [213, 214, 215, 216],
                ],
            ),
            "e": lgdo.VectorOfVectors([[1, 2, 3], [4], [], [8, 6]]),
            "ee": lgdo.VectorOfVectors([[[1], [2, 3]], [[], [4]], [[]], [[8, 6]]]),
            "tbl": lgdo.Table(
                col_dict={
                    "z": lgdo.Array([1, 1, 1, 1]),
                    "y": lgdo.Array([8, 8, 8, 8]),
                }
            ),
        }
    )

    r = obj.eval("sum(a)")
    assert isinstance(r, lgdo.Scalar)

    r = obj.eval("a + b")
    assert isinstance(r, lgdo.Array)
    assert np.array_equal(r.nda, obj.a.nda + obj.b.nda)

    r = obj.eval("a + tbl__z")
    assert isinstance(r, lgdo.Array)
    assert np.array_equal(r.nda, obj.a.nda + obj.tbl.z.nda)

    r = obj.eval("((a - b) > 1) & ((b - a) < -1)")
    assert isinstance(r, lgdo.Array)
    assert r.dtype == "bool"

    r = obj.eval("a + d")
    assert isinstance(r, lgdo.ArrayOfEqualSizedArrays)

    assert obj.eval("a**2")
    assert obj.eval("sin(a)")
    assert obj.eval("log(d)")

    r = obj.eval("a + e")
    assert isinstance(r, lgdo.VectorOfVectors)
    assert r == lgdo.VectorOfVectors([[2, 3, 4], [6], [], [12, 10]])

    r = obj.eval("2*e + 1")
    assert isinstance(r, lgdo.VectorOfVectors)
    assert r == lgdo.VectorOfVectors([[3, 5, 7], [9], [], [17, 13]])

    r = obj.eval("2*ee + 1")
    assert isinstance(r, lgdo.VectorOfVectors)
    assert r == lgdo.VectorOfVectors([[[3], [5, 7]], [[], [9]], [[]], [[17, 13]]])

    r = obj.eval("e > 2")
    assert isinstance(r, lgdo.VectorOfVectors)
    assert r == lgdo.VectorOfVectors([[False, False, True], [True], [], [True, True]])
    assert r.dtype == "bool"

    r = obj.eval("ak.sum(e, axis=-1)")
    assert isinstance(r, lgdo.Array)

    r = obj.eval("ak.sum(e)")
    assert isinstance(r, lgdo.Scalar)

    assert obj.eval("np.sum(a) + ak.sum(e)")

    # test with modules argument, the simplest is using directly lgdo
    res = obj.eval("lgdo.Array([1,2,3])", {}, modules={"lgdo": lgdo})
    assert res == lgdo.Array([1, 2, 3])

    # test with module returning np.array
    assert obj.eval("np.sum(a)", {}, modules={"np": np}).value == np.int64(10)

    # check bad type
    with pytest.raises(RuntimeError):
        obj.eval("hist.Hist()", modules={"hist": hist})

    # check impossible numexpr can still run
    assert np.allclose(
        obj.eval(
            "a*args.value",
            {"args": dbetto.AttrsDict({"value": 2})},
            modules={"lgdo": lgdo},
        ).view_as("np"),
        [2, 4, 6, 8],
    )

    # test library type selection.
    assert obj.eval("type(a) == np.ndarray").value
    assert obj.eval("type(a) == ak.highlevel.Array", library="ak").value
    assert obj.eval("type(a) == lgdo.Array", library="lgdo").value

    assert obj.eval("type(e) == ak.highlevel.Array").value
    assert obj.eval("type(e) == np.ndarray", library="np").value
    assert obj.eval("type(e) == lgdo.VectorOfVectors", library="lgdo").value



"""Implements utilities for LEGEND Data Objects."""

from __future__ import annotations

import logging
import os
from collections.abc import Iterator, MutableMapping
from typing import Any

import numpy as np

log = logging.getLogger(__name__)


def get_element_type(obj: object) -> str:
    """Get the LGDO element type of a scalar or array.

    For use in LGDO datatype attributes.

    Parameters
    ----------
    obj
        if a ``str``, will automatically return ``string`` if the object has
        a :class:`numpy.dtype`, that will be used for determining the element
        type otherwise will attempt to case the type of the object to a
        :class:`numpy.dtype`.

    Returns
    -------
    element_type
        A string stating the determined element type of the object.
    """

    # special handling for strings
    if isinstance(obj, str):
        return "string"

    # the rest use dtypes
    dt = obj.dtype if hasattr(obj, "dtype") else np.dtype(type(obj))
    kind = dt.kind

    if kind == "b":
        return "bool"
    if kind == "V":
        return "blob"
    if kind in ["i", "u", "f"]:
        return "real"
    if kind == "c":
        return "complex"
    if kind in ["S", "T", "U"]:
        return "string"
    if (
        kind == "O"
        and dt.metadata is not None
        and dt.metadata.get("vlen", None) in (str, bytes)
    ):
        # variable length strings in HDF5 are read as numpy object arrays in h5py.
        # see also h5py.check_vlen_dtype.
        return "string"

    # couldn't figure it out
    msg = "cannot determine lgdo element_type for object of type"
    raise ValueError(msg, type(obj).__name__)


def getenv_bool(name: str, default: bool = False) -> bool:
    """Get environment value as a boolean, returning True for 1, t and true
    (caps-insensitive), and False for any other value and default if undefined.
    """
    val = os.getenv(name)
    if not val:
        return default
    return val.lower() in ("1", "t", "true")


class NumbaDefaults(MutableMapping):
    """Bare-bones class to store some Numba default options. Defaults values
    are set from environment variables

    Examples
    --------
    Set all default option values for a processor at once by expanding the
    provided dictionary:

    >>> from numba import guvectorize
    >>> from lgdo.utils import numba_defaults_kwargs as nb_kwargs
    >>> @guvectorize([], "", **nb_kwargs, nopython=True) # def proc(...): ...

    Customize one argument but still set defaults for the others:

    >>> from lgdo.utils import numba_defaults as nb_defaults
    >>> @guvectorize([], "", **nb_defaults(cache=False) # def proc(...): ...

    Override global options at runtime:

    >>> from lgdo.utils import numba_defaults
    >>> # must set options before explicitly importing lgdo modules!
    >>> numba_defaults.cache = False
    >>> numba_defaults.boundscheck = True
    >>> from lgdo import compression # imports of numbified functions happen here
    >>> compression.encode(...)
    """

    def __init__(self) -> None:
        self.cache: bool = getenv_bool("LGDO_CACHE", default=True)
        self.boundscheck: bool = getenv_bool("LGDO_BOUNDSCHECK", default=False)

    def __getitem__(self, item: str) -> Any:
        return self.__dict__[item]

    def __setitem__(self, item: str, val: Any) -> None:
        self.__dict__[item] = val

    def __delitem__(self, item: str) -> None:
        del self.__dict__[item]

    def __iter__(self) -> Iterator:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __call__(self, **kwargs) -> dict:
        mapping = self.__dict__.copy()
        mapping.update(**kwargs)
        return mapping

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return str(self.__dict__)


numba_defaults = NumbaDefaults()
numba_defaults_kwargs = numba_defaults



"""
LEGEND Data Objects (LGDO) are defined in the `LEGEND data format specification
<https://github.com/legend-exp/legend-data-format-specs>`_.  This package
serves as the Python implementation of that specification. The general strategy
for the implementation is to dress standard Python and NumPy objects with an
``attr`` dictionary holding LGDO metadata, plus some convenience functions. The
basic data object classes are:

* :class:`.LGDO`: abstract base class for all LGDOs
* :class:`.Scalar`: typed Python scalar. Access data via the :attr:`value`
  attribute
* :class:`.Array`: basic :class:`numpy.ndarray`. Access data via the
  :attr:`nda` attribute.
* :class:`.FixedSizeArray`: basic :class:`numpy.ndarray`. Access data via the
  :attr:`nda` attribute.
* :class:`.ArrayOfEqualSizedArrays`: multi-dimensional :class:`numpy.ndarray`.
  Access data via the :attr:`nda` attribute.
* :class:`.VectorOfVectors`: an n-dimensional variable length array of variable
  length arrays.  Implemented as a pair of datasets: :attr:`flattened_data`
  holding the raw data (:class:`.Array` or :class:`.VectorOfVectors`, if the
  vector dimension is greater than 2), and :attr:`cumulative_length` (always an
  :class:`.Array`) whose i-th element is the sum of the lengths of the vectors
  with ``index <= i``
* :class:`.VectorOfEncodedVectors`: an array of variable length *encoded*
  arrays. Implemented as a :class:`.VectorOfVectors` :attr:`encoded_data`
  holding the encoded vectors and an :class:`.Array` :attr:`decoded_size`
  specifying the size of each decoded vector. Mainly used to represent a list
  of compressed waveforms.
* :class:`.ArrayOfEncodedEqualSizedArrays`: an array of equal sized encoded
  arrays. Similar to :class:`.VectorOfEncodedVectors` except for
  :attr:`decoded_size`, which is now a scalar.
* :class:`.Struct`: a dictionary containing LGDO objects. Derives from
  :class:`dict`
* :class:`.Table`: a :class:`.Struct` whose elements ("columns") are all array
  types with the same length (number of rows)
* :class:`.Histogram`: holds an array of histogrammed data, and the associated
  binning of arbitrary dimensionality.

Currently the primary on-disk format for LGDO object is LEGEND HDF5 (LH5) files. IO
is done via the class :class:`.lh5_store.LH5Store`. LH5 files can also be
browsed easily in python like any `HDF5 <https://www.hdfgroup.org>`_ file using
`h5py <https://www.h5py.org>`_.
"""

from __future__ import annotations

# import this so users can transparently decode data compressed with hdf5plugin
# filters
import hdf5plugin  # noqa: F401

from ._version import version as __version__
from .lh5 import LH5Iterator, ls, read, read_as, read_n_rows, show, write
from .types import (
    LGDO,
    Array,
    ArrayOfEncodedEqualSizedArrays,
    ArrayOfEqualSizedArrays,
    FixedSizeArray,
    Histogram,
    Scalar,
    Struct,
    Table,
    VectorOfEncodedVectors,
    VectorOfVectors,
    WaveformTable,
)

__all__ = [
    "LGDO",
    "Array",
    "ArrayOfEncodedEqualSizedArrays",
    "ArrayOfEqualSizedArrays",
    "FixedSizeArray",
    "Histogram",
    "LH5Iterator",
    "Scalar",
    "Struct",
    "Table",
    "VectorOfEncodedVectors",
    "VectorOfVectors",
    "WaveformTable",
    "__version__",
    "ls",
    "read",
    "read_as",
    "read_n_rows",
    "show",
    "write",
]



from __future__ import annotations

import logging

import h5py
import hdf5plugin  # noqa: F401, used for compression plugins
import numpy as np

from ....types import Scalar
from ...exceptions import LH5DecodeError
from . import utils

log = logging.getLogger(__name__)


def _h5_read_scalar(
    h5d,
    fname,
    oname,
    obj_buf=None,
):
    value = np.empty((), h5d.dtype)
    sp = h5py.h5s.create(h5py.h5s.SCALAR)
    h5d.read(sp, sp, value)
    value = value[()]
    attrs = utils.read_attrs(h5d, fname, oname)

    # special handling for bools
    # (c and Julia store as uint8 so cast to bool)
    if attrs["datatype"] == "bool":
        value = np.bool_(value)

    if obj_buf is not None:
        if not isinstance(obj_buf, Scalar):
            msg = "object buffer a Scalar"
            raise LH5DecodeError(msg, fname, oname)

        obj_buf.value = value
        obj_buf.attrs.update(attrs)
        return obj_buf, 1

    return Scalar(value=value, attrs=attrs), 1
